import json

import pytest
from indy.did import create_and_store_my_did
from indy.ledger import build_nym_request

from indy_common.authorize.auth_actions import ADD_PREFIX
from indy_common.authorize.auth_constraints import AuthConstraint, IDENTITY_OWNER
from indy_common.constants import SCHEMA, ROLE, NYM
from indy_node.test.api.helper import sdk_write_schema_and_check, sdk_build_schema_request
from indy_node.test.endorser.helper import sdk_submit_and_check_by_endorser, sdk_append_request_endorser
from indy_node.test.helper import build_auth_rule_request_json, sdk_send_and_check_req_json
from indy_node.test.nym_txn.test_create_did_without_endorser import NEW_ROLE
from plenum.common.constants import TRUSTEE, VERKEY

from plenum.common.exceptions import RequestRejectedException, RequestNackedException
from plenum.common.util import randomString
from plenum.server.request_handlers.utils import get_nym_details
from plenum.test.helper import sdk_multisign_request_object, sdk_get_and_check_replies, sdk_send_signed_requests


def test_send_by_endorser(looper, sdk_pool_handle, sdk_wallet_new_client, sdk_wallet_endorser):
    # try writing without Endorser by a None-role client
    with pytest.raises(RequestRejectedException):
        sdk_write_schema_and_check(looper, sdk_pool_handle, sdk_wallet_new_client,
                                   ["attr1", "attr2"], "name1", "1.0")

    # write with Endorser
    req_json = sdk_build_schema_request(looper, sdk_wallet_new_client,
                                        ["attr1", "attr2"], "name1", "1.0")
    sdk_submit_and_check_by_endorser(looper, sdk_pool_handle,
                                     sdk_wallet_author=sdk_wallet_new_client, sdk_wallet_endorser=sdk_wallet_endorser,
                                     request_json=req_json)


def test_both_author_and_endorser_must_sign(looper, sdk_pool_handle, sdk_wallet_trustee, sdk_wallet_endorser):
    '''
    Both author and endorser must sign the request even if the author can send the request without Endorser
    '''
    req_json = sdk_build_schema_request(looper, sdk_wallet_trustee,
                                        ["attr1", "attr2"], "name1", "2.0")
    req_json = sdk_append_request_endorser(looper, req_json, sdk_wallet_endorser[1])

    # sign by Author only
    req_json_author_only = sdk_multisign_request_object(looper, sdk_wallet_trustee, req_json)
    with pytest.raises(RequestNackedException, match="Endorser must sign the request"):
        request_couple = sdk_send_signed_requests(sdk_pool_handle, [req_json_author_only])[0]
        sdk_get_and_check_replies(looper, [request_couple])

    # sign by Endorser only
    req_json_endorser_only = sdk_multisign_request_object(looper, sdk_wallet_endorser, req_json)
    with pytest.raises(RequestNackedException, match="Author must sign the request when sending via Endorser"):
        request_couple = sdk_send_signed_requests(sdk_pool_handle, [req_json_endorser_only])[0]
        sdk_get_and_check_replies(looper, [request_couple])

    # sign by both
    req_json_both = sdk_multisign_request_object(looper, sdk_wallet_trustee, req_json)
    req_json_both = sdk_multisign_request_object(looper, sdk_wallet_endorser, req_json_both)
    request_couple = sdk_send_signed_requests(sdk_pool_handle, [req_json_both])[0]
    sdk_get_and_check_replies(looper, [request_couple])


def test_endorser_field_must_be_explicit(looper, sdk_pool_handle, sdk_wallet_new_client, sdk_wallet_endorser):
    req_json = sdk_build_schema_request(looper, sdk_wallet_new_client,
                                        ["attr1", "attr2"], "name2", "3.0")
    # Do not append Endorser field!

    # sign by both Author and Endorser
    req_json = sdk_multisign_request_object(looper, sdk_wallet_new_client, req_json)
    req_json = sdk_multisign_request_object(looper, sdk_wallet_endorser, req_json)

    with pytest.raises(RequestRejectedException,
                       match="'Endorser' field must be explicitly set for the endorsed transaction"):
        request_couple = sdk_send_signed_requests(sdk_pool_handle, [req_json])[0]
        sdk_get_and_check_replies(looper, [request_couple])


def test_endorser_must_have_known_role(looper, sdk_pool_handle, sdk_wallet_new_client, sdk_wallet_trustee):
    req_json = sdk_build_schema_request(looper, sdk_wallet_trustee,
                                        ["attr1", "attr2"], "name3", "4.0")
    with pytest.raises(RequestRejectedException, match="Endorser must have one of the following roles"):
        sdk_submit_and_check_by_endorser(looper, sdk_pool_handle,
                                         sdk_wallet_author=sdk_wallet_trustee,
                                         sdk_wallet_endorser=sdk_wallet_new_client,
                                         request_json=req_json)


def test_endorser_not_required_when_two_trustee_sigs(looper, sdk_pool_handle, sdk_wallet_trustee_list):
    change_new_schema_auth_rule(looper, sdk_pool_handle, sdk_wallet_trustee_list[0],
                                constraint=AuthConstraint(role=TRUSTEE, sig_count=2))

    req_json = sdk_build_schema_request(looper, sdk_wallet_trustee_list[1],
                                        ["attr1", "attr2"], "name4", "5.0")
    # sign by both
    req_json = sdk_multisign_request_object(looper, sdk_wallet_trustee_list[1], req_json)
    req_json = sdk_multisign_request_object(looper, sdk_wallet_trustee_list[2], req_json)
    request_couple = sdk_send_signed_requests(sdk_pool_handle, [req_json])[0]
    sdk_get_and_check_replies(looper, [request_couple])


def test_endorser_required_when_multi_sig_with_off_ledger_signature(looper, txnPoolNodeSet,
                                                                    sdk_wallet_client,
                                                                    sdk_pool_handle,
                                                                    sdk_wallet_trustee,
                                                                    sdk_wallet_endorser):
    change_new_did_auth_rule(looper, sdk_pool_handle, sdk_wallet_trustee,
                             constraint=AuthConstraint(role='*',
                                                       sig_count=2,
                                                       off_ledger_signature=True))

    new_did, verkey = \
        looper.loop.run_until_complete(
            create_and_store_my_did(sdk_wallet_client[0], json.dumps({'seed': randomString(32)})))
    sdk_wallet_new_client = (sdk_wallet_client[0], new_did)
    nym_request = looper.loop.run_until_complete(
        build_nym_request(new_did, new_did, verkey, None, IDENTITY_OWNER))

    # can not send without Endorser
    with pytest.raises(RequestRejectedException,
                       match="'Endorser' field must be explicitly set for the endorsed transaction"):
        signed_req = sdk_multisign_request_object(looper, sdk_wallet_new_client, nym_request)
        signed_req = sdk_multisign_request_object(looper, sdk_wallet_endorser, signed_req)
        request_couple = sdk_send_signed_requests(sdk_pool_handle, [signed_req])[0]
        sdk_get_and_check_replies(looper, [request_couple])

    # can send with Endorser
    sdk_submit_and_check_by_endorser(looper, sdk_pool_handle,
                                     sdk_wallet_author=sdk_wallet_new_client, sdk_wallet_endorser=sdk_wallet_endorser,
                                     request_json=nym_request)
    details = get_nym_details(txnPoolNodeSet[0].states[1], new_did, is_committed=True)
    assert details[ROLE] == IDENTITY_OWNER
    assert details[VERKEY] == verkey


def change_new_schema_auth_rule(looper, sdk_pool_handle, sdk_wallet_trustee, constraint):
    req = build_auth_rule_request_json(
        looper, sdk_wallet_trustee[1],
        auth_action=ADD_PREFIX,
        auth_type=SCHEMA,
        field='*',
        old_value=None,
        new_value='*',
        constraint=constraint.as_dict
    )

    sdk_send_and_check_req_json(looper, sdk_pool_handle, sdk_wallet_trustee, req)


def change_new_did_auth_rule(looper, sdk_pool_handle, sdk_wallet_trustee, constraint):
    req = build_auth_rule_request_json(
        looper, sdk_wallet_trustee[1],
        auth_action=ADD_PREFIX,
        auth_type=NYM,
        field=ROLE,
        old_value=None,
        new_value=NEW_ROLE,
        constraint=constraint.as_dict
    )

    sdk_send_and_check_req_json(looper, sdk_pool_handle, sdk_wallet_trustee, req)



import pytest
import typing

from indy_common.strict_types import strict_types, decClassMethods


@strict_types()
def takesStr(s: str) -> int:
    try:
        return int(s)
    except ValueError:
        pass


def testInvalidArgumentType():
    with pytest.raises(TypeError):
        takesStr(1)


def testInvalidReturnType():
    with pytest.raises(TypeError):
        assert takesStr('return None')


def testValidInputAndReturn():
    takesStr('1')


@decClassMethods(strict_types())
class TestClass:

    def takesStr(self, s: str) -> int:
        try:
            return int(s)
        except ValueError:
            pass

    def takesInt(self, i: int) -> str:
        try:
            if i != 457:
                return str(i)
        except ValueError:
            pass


@pytest.fixture(scope="module")
def t():
    return TestClass()


def testWholeClassInvalidArgumentType(t):
    with pytest.raises(TypeError):
        t.takesStr(1)
    with pytest.raises(TypeError):
        t.takesInt('1')


def testWholeClassInvalidReturnType(t):
    with pytest.raises(TypeError):
        assert t.takesStr('return None')
        assert t.takesInt(457)


def testWholeClassValidInputAndReturn(t):
    t.takesStr('1')
    t.takesInt(1)



import pytest

from common.serializers.serialization import config_state_serializer
from indy_common.authorize.auth_actions import ADD_PREFIX, AuthActionAdd
from indy_common.authorize.auth_constraints import AuthConstraint, ConstraintsSerializer
from indy_common.constants import NYM, CONFIG_LEDGER_ID
from indy_common.state import config
from indy_node.test.auth_rule.helper import create_verkey_did, sdk_send_and_check_auth_rule_request
from plenum.common.constants import STEWARD, ROLE, STEWARD_STRING
from plenum.common.exceptions import RequestRejectedException
from plenum.common.startable import Mode
from plenum.test.delayers import cDelay, ppDelay, pDelay
from plenum.test.helper import assertExp
from plenum.test.node_catchup.helper import ensure_all_nodes_have_same_data
from plenum.test.pool_transactions.helper import sdk_add_new_nym
from plenum.test.stasher import delay_rules_without_processing
from stp_core.loop.eventually import eventually


def test_catching_up_auth_rule_txn(looper,
                                   txnPoolNodeSet,
                                   sdk_wallet_trustee,
                                   sdk_wallet_steward,
                                   sdk_pool_handle):
    delayed_node = txnPoolNodeSet[-1]
    wh, _ = sdk_wallet_trustee
    new_steward_did, new_steward_verkey = create_verkey_did(looper, wh)
    changed_constraint = AuthConstraint(role=STEWARD,
                                        sig_count=1)
    action = AuthActionAdd(txn_type=NYM,
                           field=ROLE,
                           value=STEWARD)
    with pytest.raises(RequestRejectedException, match="Not enough TRUSTEE signatures"):
        sdk_add_new_nym(looper,
                        sdk_pool_handle,
                        sdk_wallet_steward,
                        'newSteward2',
                        STEWARD_STRING,
                        dest=new_steward_did, verkey=new_steward_verkey)
    with delay_rules_without_processing(delayed_node.nodeIbStasher, cDelay(), pDelay(), ppDelay()):
        sdk_send_and_check_auth_rule_request(looper, sdk_pool_handle, sdk_wallet_trustee,
                                             auth_action=ADD_PREFIX,
                                             auth_type=action.txn_type, field=action.field,
                                             new_value=action.value, old_value=None,
                                             constraint=changed_constraint.as_dict)
        sdk_add_new_nym(looper,
                        sdk_pool_handle,
                        sdk_wallet_trustee,
                        'newSteward2')
        delayed_node.start_catchup()
        looper.run(eventually(lambda: assertExp(delayed_node.mode == Mode.participating)))
    sdk_add_new_nym(looper,
                    sdk_pool_handle,
                    sdk_wallet_steward,
                    'newSteward3',
                    STEWARD_STRING,
                    dest=new_steward_did, verkey=new_steward_verkey)
    ensure_all_nodes_have_same_data(looper, txnPoolNodeSet)
    config_state = delayed_node.states[CONFIG_LEDGER_ID]
    from_state = config_state.get(config.make_state_path_for_auth_rule(action.get_action_id()),
                                  isCommitted=True)
    assert changed_constraint == ConstraintsSerializer(config_state_serializer).deserialize(from_state)



#!/usr/bin/env python3

import shutil
import argparse
from indy_common.config_util import getConfig
from indy_common.config_helper import ConfigHelper


def clean(config, full, network_name):
    if network_name:
        config.NETWORK_NAME = network_name
    config_helper = ConfigHelper(config)

    shutil.rmtree(config_helper.log_dir)
    shutil.rmtree(config_helper.keys_dir)
    shutil.rmtree(config_helper.genesis_dir)

    if full:
        shutil.rmtree(config_helper.ledger_base_dir)
        shutil.rmtree(config_helper.log_base_dir)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='Removes node data and configuration')
    parser.add_argument('--full', required=False, type=bool, default=False, help="remove configs and logs too")
    parser.add_argument('--network', required=False, type=str, help="Network to clean")
    args = parser.parse_args()
    config = getConfig()
    clean(config, args.full, args.network)



from indy_common.constants import TIMESTAMP, REVOC_REG_DEF_ID, GET_REVOC_REG
from indy_node.server.request_handlers.domain_req_handlers.revoc_reg_entry_handler import RevocRegEntryHandler
from indy_node.server.request_handlers.utils import StateValue
from plenum.common.constants import DOMAIN_LEDGER_ID
from plenum.common.request import Request
from plenum.server.database_manager import DatabaseManager
from plenum.server.request_handlers.handler_interfaces.read_request_handler import ReadRequestHandler
from plenum.server.request_handlers.utils import decode_state_value


class GetRevocRegHandler(ReadRequestHandler):

    def __init__(self, database_manager: DatabaseManager):
        super().__init__(database_manager, GET_REVOC_REG, DOMAIN_LEDGER_ID)

    def get_result(self, request: Request):
        self._validate_request_type(request)
        req_ts = request.operation.get(TIMESTAMP)
        revoc_reg_def_id = request.operation.get(REVOC_REG_DEF_ID)
        # Get root hash corresponding with given timestamp
        past_root = self.database_manager.ts_store.get_equal_or_prev(req_ts)
        # Path to corresponding ACCUM record in state
        path = RevocRegEntryHandler.make_state_path_for_revoc_reg_entry_accum(revoc_reg_def_id=revoc_reg_def_id)
        entry_state = StateValue()
        if past_root is not None:
            encoded_entry, proof = self._get_value_from_state(path,
                                                              head_hash=past_root,
                                                              with_proof=True)
            entry_state.proof = proof
            if encoded_entry:
                revoc_reg_entry_accum, seq_no, last_update_time = decode_state_value(encoded_entry)
                entry_state = StateValue(root_hash=past_root,
                                         value=revoc_reg_entry_accum,
                                         seq_no=seq_no,
                                         update_time=last_update_time,
                                         proof=proof)

        return self.make_result(request=request,
                                data=entry_state.value,
                                last_seq_no=entry_state.seq_no,
                                update_time=entry_state.update_time,
                                proof=entry_state.proof)

    def _get_revoc_def_entry(self,
                             revoc_reg_def_id,
                             is_committed=True) -> (str, int, int, list):
        assert revoc_reg_def_id
        path = RevocRegEntryHandler.make_state_path_for_revoc_reg_entry(revoc_reg_def_id=revoc_reg_def_id)
        try:
            keys, seq_no, last_update_time, proof = self.lookup(path, is_committed, with_proof=True)
            return keys, seq_no, last_update_time, proof
        except KeyError:
            return None, None, None, None



import rlp

from indy_common.constants import ROLE, ENDORSER
from plenum.common.constants import VERKEY, TRUSTEE, STEWARD, THREE_PC_PREFIX, \
    TXN_TIME
from plenum.common.types import f
from storage.kv_store import KeyValueStorage
from storage.optimistic_kv_store import OptimisticKVStore
from stp_core.common.log import getlogger

logger = getlogger()


class IdrCache(OptimisticKVStore):
    """
    A cache to store a role and verkey of an identifier, the db is only used to
    store committed data, uncommitted data goes to memory
    The key is the identifier and value is a pack of fields in rlp
    The first item is the endorser, the second item is verkey and the
    third item is role
    """

    unsetVerkey = b'-'

    def __init__(self, name, keyValueStorage: KeyValueStorage):
        logger.debug('Initializing identity cache {}'.format(name))
        self._keyValueStorage = keyValueStorage
        super().__init__(self._keyValueStorage)
        self._name = name

    def __repr__(self):
        return self._name

    @staticmethod
    def encodeVerkey(verkey):
        if verkey is None:
            return IdrCache.unsetVerkey
        if isinstance(verkey, str):
            return verkey.encode()
        return verkey

    @staticmethod
    def decodeVerkey(verkey):
        if verkey != IdrCache.unsetVerkey:
            return verkey.decode()

    @staticmethod
    def packIdrValue(seqNo, txnTime, ta, role, verkey):
        if seqNo is None:
            raise ValueError("seqNo should not be None!")
        seqNo = str(seqNo)
        txnTime = b'' if txnTime is None else str(txnTime)
        if ta is None:
            ta = b''
        if role is None:
            role = b''
        verkey = IdrCache.encodeVerkey(verkey)
        return rlp.encode([seqNo, txnTime, ta, role, verkey])

    @staticmethod
    def unpackIdrValue(value):
        if value is None:
            return None
        seqNo, txnTime, ta, role, verkey = rlp.decode(value)
        seqNo = int(seqNo.decode())
        txnTime = txnTime.decode()
        txnTime = int(txnTime) if txnTime else None
        ta = ta.decode()
        role = role.decode()
        verkey = IdrCache.decodeVerkey(verkey)
        return seqNo, txnTime, ta, role, verkey

    def get(self, idr, isCommitted=True):
        idr = idr.encode()
        value = super().get(idr, is_committed=isCommitted)
        return self.unpackIdrValue(value)

    def set(self, idr, seqNo, txnTime,
            ta=None, role=None, verkey=None, isCommitted=True):
        idr = idr.encode()
        val = self.packIdrValue(seqNo, txnTime, ta, role, verkey)
        super().set(idr, val, is_committed=isCommitted)

    def close(self):
        self._keyValueStorage.close()

    def currentBatchCreated(self, stateRoot, ppTime):
        super().create_batch_from_current(stateRoot)

    def batchRejected(self):
        super().reject_batch()

    def onBatchCommitted(self, stateRoot):
        # Commit an already created batch
        batch_idr = super().first_batch_idr
        if batch_idr is None:
            logger.warning('{}{} is trying to commit a batch with state root'
                           ' {} but no uncommitted found'
                           .format(THREE_PC_PREFIX, self, stateRoot))
            return
        if batch_idr != stateRoot:
            logger.warning('{}{}: The first created batch has not been '
                           'committed or reverted and yet another batch is '
                           'trying to be committed, {} {}'.
                           format(THREE_PC_PREFIX, self,
                                  self.un_committed[0][0], stateRoot))
            return

        try:
            return super().commit_batch()
        except ValueError:
            logger.warning('{}{} found no uncommitted batch'.
                           format(THREE_PC_PREFIX, self))

    def getVerkey(self, idr, isCommitted=True):
        seqNo, txnTime, ta, role, verkey = self.get(idr, isCommitted=isCommitted)
        return verkey

    def getRole(self, idr, isCommitted=True):
        seqNo, txnTime, ta, role, verkey = self.get(idr, isCommitted=isCommitted)
        return role

    def getNym(self, nym, role=None, isCommitted=True):
        """
        Get a nym, if role is provided then get nym with that role
        :param nym:
        :param role:
        :param isCommitted:
        :return:
        """
        try:
            seqNo, txnTime, ta, actual_role, verkey = self.get(nym, isCommitted)
        except KeyError:
            return None
        if role and role != actual_role:
            return None
        return {
            ROLE: actual_role or None,
            VERKEY: verkey or None,
            f.IDENTIFIER.nm: ta or None,
            f.SEQ_NO.nm: seqNo or None,
            TXN_TIME: txnTime or None,
        }

    def getTrustee(self, nym, isCommitted=True):
        return self.getNym(nym, TRUSTEE, isCommitted=isCommitted)

    def getSteward(self, nym, isCommitted=True):
        return self.getNym(nym, STEWARD, isCommitted=isCommitted)

    def getEndorser(self, nym, isCommitted=True):
        return self.getNym(nym, ENDORSER, isCommitted=isCommitted)

    def hasTrustee(self, nym, isCommitted=True):
        return bool(self.getTrustee(nym, isCommitted=isCommitted))

    def hasSteward(self, nym, isCommitted=True):
        return bool(self.getSteward(nym, isCommitted=isCommitted))

    def hasEndorser(self, nym, isCommitted=True):
        return bool(self.getEndorser(nym, isCommitted=isCommitted))

    def hasNym(self, nym, isCommitted=True):
        return bool(self.getNym(nym, isCommitted=isCommitted))

    def getOwnerFor(self, nym, isCommitted=True):
        nymData = self.getNym(nym, isCommitted=isCommitted)
        if nymData:
            if nymData.get(VERKEY) is None:
                return nymData[f.IDENTIFIER.nm]
            return nym
        logger.info('Nym {} not found'.format(nym))



import pytest
from indy import did

from plenum.common.exceptions import RequestRejectedException
from plenum.common.constants import TRUSTEE_STRING
from plenum.test.pool_transactions.helper import sdk_add_new_nym
from plenum.test.helper import sdk_get_and_check_replies, sdk_sign_and_submit_op


def test_new_DID_cannot_update_another_DID(looper,
                                           sdk_pool_handle,
                                           sdk_wallet_trustee,
                                           sdk_wallet_handle):
    """Create trustee"""
    trustee_did, trustee_verkey = looper.loop.run_until_complete(
        did.create_and_store_my_did(sdk_wallet_trustee[0], "{}"))

    """Add trustee to ledger"""
    sdk_add_new_nym(looper, sdk_pool_handle,
                    sdk_wallet_trustee, 'newTrustee', TRUSTEE_STRING, verkey=trustee_verkey, dest=trustee_did)

    """new DID (no role)"""
    new_no_role_did, new_no_role_verkey = looper.loop.run_until_complete(
        did.create_and_store_my_did(sdk_wallet_trustee[0], "{}"))

    """Adding new DID (no role) to ledger"""
    sdk_add_new_nym(looper, sdk_pool_handle,
                    sdk_wallet_trustee,
                    'noRole',
                    verkey=new_no_role_verkey, dest=new_no_role_did)

    """Nym transaction to update Trustee DID that makes no change to verkey or role"""
    op = {'type': '1',
          'dest': trustee_did
          }

    """Submitting the transaction fails"""
    with pytest.raises(RequestRejectedException):
        req = sdk_sign_and_submit_op(looper, sdk_pool_handle, sdk_wallet_trustee, op)
        sdk_get_and_check_replies(looper, [req])



import importlib
import time
import os

from indy_node.__metadata__ import __version__ as node_pgk_version
from plenum.server.validator_info_tool import none_on_fail, \
    ValidatorNodeInfoTool as PlenumValidatorNodeInfoTool
from plenum.common.constants import POOL_LEDGER_ID, DOMAIN_LEDGER_ID, CONFIG_LEDGER_ID, AUDIT_LEDGER_ID


class ValidatorNodeInfoTool(PlenumValidatorNodeInfoTool):

    @property
    def info(self):
        info = super().info
        ts_str = "{}".format(time.strftime(
            "%A, %B %{0}d, %Y %{0}I:%M:%S %p %z".format('#' if os.name == 'nt' else '-'),
            time.localtime(info["timestamp"])))
        info.update({"Update_time": ts_str})
        if 'Node_info' in info:
            if 'Metrics' in info['Node_info']:
                std_ledgers = [POOL_LEDGER_ID, DOMAIN_LEDGER_ID,
                               CONFIG_LEDGER_ID, AUDIT_LEDGER_ID]
                other_ledgers = {}
                for idx, linfo in self._node.ledgerManager.ledgerRegistry.items():
                    if linfo.id in std_ledgers:
                        continue
                    other_ledgers[linfo.id] = linfo.ledger.size
                info['Node_info']['Metrics']['transaction-count'].update(other_ledgers)

        return info

    @none_on_fail
    def _generate_software_info(self):
        sfv = super()._generate_software_info()
        sfv['Software'].update({'indy-node': self.__node_pkg_version})
        sfv['Software'].update({'sovrin': "unknown"})
        try:
            pkg = importlib.import_module(self._config.UPGRADE_ENTRY)
            sfv['Software'].update({self._config.UPGRADE_ENTRY: pkg.__version__})
        except Exception:
            pass

        return sfv

    @property
    @none_on_fail
    def __node_pkg_version(self):
        return node_pgk_version



from typing import NamedTuple

from plenum.common.constants import TXN_TYPE, TARGET_NYM, ORIGIN, DATA, RAW, \
    ENC, HASH, NAME, VERSION, ALIAS, VERKEY, FORCE
from indy_common.roles import Roles
from indy_common.transactions import IndyTransactions

Environment = NamedTuple("Environment", [
    ("poolLedger", str),
    ("domainLedger", str)
])

DOMAIN_LEDGER_ID = 1

# NYM
DIDDOC_CONTENT = "diddocContent"
NYM_VERSION = "version"
VERSION_ID = "seqNo"
VERSION_TIME = "timestamp"

NYM_VERSION_NULL = 0  # No Checks
NYM_VERSION_CONVENTION = 1  # Nym is base58 of first 16 bytes of verkey
NYM_VERSION_SELF_CERT = 2  # Nym is base58 of first 16 bytes of sha256 of verkey
DEFAULT_NYM_VERSION = NYM_VERSION_NULL  # No checks by default

# SCHEMA
SCHEMA_NAME = "name"
SCHEMA_VERSION = "version"
SCHEMA_ATTR_NAMES = "attr_names"
SCHEMA_FROM = "dest"

# CLAIM DEF
CLAIM_DEF_SIGNATURE_TYPE = "signature_type"
CLAIM_DEF_SCHEMA_REF = "ref"
CLAIM_DEF_TAG = "tag"
CLAIM_DEF_PUBLIC_KEYS = "data"
CLAIM_DEF_FROM = "origin"
CLAIM_DEF_PRIMARY = "primary"
CLAIM_DEF_REVOCATION = "revocation"
CLAIM_DEF_TAG_DEFAULT = "tag"
CLAIM_DEF_CL = "CL"

# AUTH_RULE
CONSTRAINT = "constraint"
OLD_VALUE = "old_value"
NEW_VALUE = "new_value"
AUTH_ACTION = "auth_action"
AUTH_TYPE = "auth_type"
FIELD = "field"
RULES = "rules"

ROLE = 'role'
NONCE = 'nonce'
ATTRIBUTES = "attributes"
ACTION = 'action'
SCHEDULE = 'schedule'
DATETIME = 'datetime'
TIMEOUT = 'timeout'
SHA256 = 'sha256'
START = 'start'
CANCEL = 'cancel'
COMPLETE = 'complete'
IN_PROGRESS = 'in_progress'
FAIL = 'fail'
JUSTIFICATION = 'justification'
REINSTALL = 'reinstall'
SIGNATURE_TYPE = 'signature_type'
TAG = 'tag'
PACKAGE = 'package'

REVOC_TYPE = "revocDefType"
ID = "id"
CRED_DEF_ID = "credDefId"
ISSUANCE_TYPE = "issuanceType"
MAX_CRED_NUM = "maxCredNum"
PUBLIC_KEYS = "publicKeys"
TAILS_HASH = "tailsHash"
TAILS_LOCATION = "tailsLocation"
SUBMITTER_DID = "submitterDid"
VALUE = "value"
REQ_METADATA = "reqMetadata"
REVOC_REG_DEF_ID = "revocRegDefId"
PREV_ACCUM = "prevAccum"
ACCUM = "accum"
ACCUM_FROM = "accum_from"
ACCUM_TO = "accum_to"
ISSUED = "issued"
REVOKED = "revoked"
ISSUANCE_BY_DEFAULT = "ISSUANCE_BY_DEFAULT"
ISSUANCE_ON_DEMAND = "ISSUANCE_ON_DEMAND"
TIMESTAMP = 'timestamp'
FROM = "from"
TO = "to"
STATE_PROOF_FROM = "stateProofFrom"
REVOC_REG_ID = "revocRegId"

NULL = 'null'
OWNER = '<owner>'

LAST_TXN = "lastTxn"
TXNS = "Txns"

ENC_TYPE = "encType"
SKEY = "secretKey"
REF = "ref"
PRIMARY = "primary"
REVOCATION = "revocation"

WRITES = "writes"

RESTART_MESSAGE = "restart"
UPGRADE_MESSAGE = "upgrade"
MESSAGE_TYPE = "message_type"

allOpKeys = (
    TXN_TYPE,
    TARGET_NYM,
    VERKEY,
    ORIGIN,
    ROLE,
    DATA,
    NONCE,
    REF,
    RAW,
    ENC,
    HASH,
    ALIAS,
    ACTION,
    SCHEDULE,
    TIMEOUT,
    SHA256,
    START,
    CANCEL,
    NAME,
    VERSION,
    JUSTIFICATION,
    SIGNATURE_TYPE,
    FORCE,
    WRITES,
    REINSTALL)

reqOpKeys = (TXN_TYPE,)

# Attribute Names
ENDPOINT = "endpoint"

# Roles

# FIXME can be automated by iteration through Roles
# but it would be less self-descriptive
ENDORSER = Roles.ENDORSER.value
ENDORSER_STRING = 'ENDORSER'

NETWORK_MONITOR = Roles.NETWORK_MONITOR.value
NETWORK_MONITOR_STRING = 'NETWORK_MONITOR'

# client transaction types
NODE = IndyTransactions.NODE.value
NYM = IndyTransactions.NYM.value
ATTRIB = IndyTransactions.ATTRIB.value
SCHEMA = IndyTransactions.SCHEMA.value
CLAIM_DEF = IndyTransactions.CLAIM_DEF.value
REVOC_REG_DEF = IndyTransactions.REVOC_REG_DEF.value
REVOC_REG_ENTRY = IndyTransactions.REVOC_REG_ENTRY.value
DISCLO = IndyTransactions.DISCLO.value
GET_ATTR = IndyTransactions.GET_ATTR.value
GET_NYM = IndyTransactions.GET_NYM.value
GET_TXNS = IndyTransactions.GET_TXNS.value
GET_SCHEMA = IndyTransactions.GET_SCHEMA.value
GET_CLAIM_DEF = IndyTransactions.GET_CLAIM_DEF.value
GET_REVOC_REG_DEF = IndyTransactions.GET_REVOC_REG_DEF.value
GET_REVOC_REG = IndyTransactions.GET_REVOC_REG.value
GET_REVOC_REG_DELTA = IndyTransactions.GET_REVOC_REG_DELTA.value
CHANGE_KEY = IndyTransactions.CHANGE_KEY.value

POOL_UPGRADE = IndyTransactions.POOL_UPGRADE.value
NODE_UPGRADE = IndyTransactions.NODE_UPGRADE.value
POOL_RESTART = IndyTransactions.POOL_RESTART.value
VALIDATOR_INFO = IndyTransactions.VALIDATOR_INFO.value

POOL_CONFIG = IndyTransactions.POOL_CONFIG.value
AUTH_RULE = IndyTransactions.AUTH_RULE.value
AUTH_RULES = IndyTransactions.AUTH_RULES.value
GET_AUTH_RULE = IndyTransactions.GET_AUTH_RULE.value

# client Rich Schema transaction types
JSON_LD_CONTEXT = IndyTransactions.JSON_LD_CONTEXT.value
RICH_SCHEMA = IndyTransactions.RICH_SCHEMA.value
RICH_SCHEMA_ENCODING = IndyTransactions.RICH_SCHEMA_ENCODING.value
RICH_SCHEMA_MAPPING = IndyTransactions.RICH_SCHEMA_MAPPING.value
RICH_SCHEMA_CRED_DEF = IndyTransactions.RICH_SCHEMA_CRED_DEF.value
RICH_SCHEMA_PRES_DEF = IndyTransactions.RICH_SCHEMA_PRES_DEF.value
GET_RICH_SCHEMA_OBJECT_BY_ID = IndyTransactions.GET_RICH_SCHEMA_OBJECT_BY_ID.value
GET_RICH_SCHEMA_OBJECT_BY_METADATA = IndyTransactions.GET_RICH_SCHEMA_OBJECT_BY_METADATA.value

CONFIG_LEDGER_ID = 2
JUSTIFICATION_MAX_SIZE = 1000

LOCAL_AUTH_POLICY = 1
CONFIG_LEDGER_AUTH_POLICY = 2

TAG_LIMIT_SIZE = 256

APP_NAME = "indy-node"

# RICH SCHEMA
RS_ID = 'id'
RS_TYPE = 'rsType'
RS_NAME = 'rsName'
RS_VERSION = 'rsVersion'
RS_CONTENT = 'content'
RS_FROM = 'from'
RS_ENDORSER = 'endorser'

JSON_LD_CONTEXT_FIELD = "@context"
JSON_LD_ID_FIELD = "@id"
JSON_LD_TYPE_FIELD = "@type"

# RICH SCHEMA type names
RS_CONTEXT_TYPE_VALUE = 'ctx'
RS_SCHEMA_TYPE_VALUE = 'sch'
RS_ENCODING_TYPE_VALUE = 'enc'
RS_MAPPING_TYPE_VALUE = 'map'
RS_CRED_DEF_TYPE_VALUE = 'cdf'
RS_PRES_DEF_TYPE_VALUE = 'pdf'

# Specific Rich Schema Object's content fields:
RS_CRED_DEF_SIG_TYPE = "signatureType"
RS_CRED_DEF_MAPPING = "mapping"
RS_CRED_DEF_SCHEMA = "schema"
RS_CRED_DEF_PUB_KEY = "publicKey"

RS_ENC_INPUT = "input"
RS_ENC_OUTPUT = "output"
RS_ENC_ALGORITHM = "algorithm"
RS_ENC_TEST_VECS = "testVectors"
RS_ENC_ID = "id"
RS_ENC_TYPE = "type"
RS_ENC_ALG_DESC = "description"
RS_ENC_ALG_DOC = "documentation"
RS_ENC_ALG_IMPL = "implementation"

RS_MAPPING_ATTRIBUTES = "attributes"
RS_MAPPING_ISSUER = "issuer"
RS_MAPPING_ISSUANCE_DATE = "issuanceDate"
RS_MAPPING_SCHEMA = "schema"
RS_MAPPING_ENC = "enc"
RS_MAPPING_RANK = "rank"



import json
import time

import pytest
from plenum.common.exceptions import RequestNackedException

from indy_common.constants import NODE_UPGRADE
from plenum.test.helper import sdk_get_and_check_replies
from plenum.test.pool_transactions.helper import sdk_sign_and_send_prepared_request

from plenum.common.constants import TXN_TYPE, CURRENT_PROTOCOL_VERSION
from plenum.common.types import OPERATION, f


def test_send_node_upgrade_txn(looper, sdk_wallet_client, sdk_pool_handle):
    req = {
        OPERATION: {
            TXN_TYPE: NODE_UPGRADE,
            'data': 'data1'
        },
        f.IDENTIFIER.nm: sdk_wallet_client[1],
        f.REQ_ID.nm: int(time.time()),
        f.PROTOCOL_VERSION.nm: CURRENT_PROTOCOL_VERSION
    }

    rep = sdk_sign_and_send_prepared_request(looper, sdk_wallet_client, sdk_pool_handle, json.dumps(req))
    with pytest.raises(RequestNackedException):
        sdk_get_and_check_replies(looper, [rep])



import pytest
from indy import did

from indy_common.constants import ENDORSER_STRING
from plenum.common.constants import TRUSTEE_STRING, STEWARD_STRING
from plenum.common.exceptions import RequestRejectedException
from plenum.test.helper import sdk_get_and_check_replies, sdk_sign_and_submit_op
from plenum.test.pool_transactions.helper import sdk_add_new_nym


def test_steward_suspension_by_another_trustee(looper,
                                               sdk_pool_handle,
                                               sdk_wallet_trustee,
                                               sdk_wallet_handle,
                                               with_verkey):
    new_trustee_did, new_trustee_verkey = looper.loop.run_until_complete(
        did.create_and_store_my_did(sdk_wallet_trustee[0], "{}"))
    new_steward_did, new_steward_verkey = looper.loop.run_until_complete(
        did.create_and_store_my_did(sdk_wallet_trustee[0], "{}"))

    """Adding new steward"""
    sdk_add_new_nym(looper, sdk_pool_handle,
                    sdk_wallet_trustee, 'newSteward', STEWARD_STRING, verkey=new_steward_verkey, dest=new_steward_did)

    """Adding new trustee"""
    sdk_add_new_nym(looper, sdk_pool_handle,
                    sdk_wallet_trustee, 'newTrustee', TRUSTEE_STRING, verkey=new_trustee_verkey, dest=new_trustee_did)

    """Blacklisting new steward by new trustee"""
    op = {'type': '1',
          'dest': new_steward_did,
          'role': None}
    if with_verkey:
        op['verkey'] = new_steward_verkey
    req = sdk_sign_and_submit_op(looper, sdk_pool_handle, (sdk_wallet_handle, new_trustee_did), op)
    if with_verkey:
        with pytest.raises(RequestRejectedException):
            sdk_get_and_check_replies(looper, [req])
    else:
        sdk_get_and_check_replies(looper, [req])


def test_steward_cannot_create_endorsers_after_demote(looper,
                                                          sdk_pool_handle,
                                                          sdk_wallet_trustee,
                                                          sdk_wallet_handle):
    new_steward_did, new_steward_verkey = looper.loop.run_until_complete(
        did.create_and_store_my_did(sdk_wallet_trustee[0], "{}"))
    new_ta_did, new_ta_verkey = looper.loop.run_until_complete(
        did.create_and_store_my_did(sdk_wallet_trustee[0], "{}"))
    new_ta_2_did, new_ta_2_verkey = looper.loop.run_until_complete(
        did.create_and_store_my_did(sdk_wallet_trustee[0], "{}"))

    """Adding new steward"""
    sdk_add_new_nym(looper, sdk_pool_handle,
                    sdk_wallet_trustee,
                    'newSteward',
                    STEWARD_STRING,
                    verkey=new_steward_verkey, dest=new_steward_did)

    """Adding new TA"""
    sdk_add_new_nym(looper, sdk_pool_handle,
                    (sdk_wallet_handle, new_steward_did),
                    'newSteward',
                    ENDORSER_STRING,
                    verkey=new_ta_verkey, dest=new_ta_did)

    """Blacklisting new steward by trustee"""
    op = {'type': '1',
          'dest': new_steward_did,
          'role': None}
    req = sdk_sign_and_submit_op(looper, sdk_pool_handle, sdk_wallet_trustee, op)
    sdk_get_and_check_replies(looper, [req])

    """Try to add new TA by previous demoted steward"""
    with pytest.raises(RequestRejectedException):
        sdk_add_new_nym(looper, sdk_pool_handle,
                        (sdk_wallet_handle, new_steward_did),
                        'newSteward',
                        ENDORSER_STRING,
                        verkey=new_ta_2_verkey, dest=new_ta_2_did)



import json

from indy.ledger import build_attrib_request
from indy_node.test.api.helper import validate_write_reply, validate_attrib_txn
from plenum.test.helper import sdk_get_reply, sdk_sign_and_submit_req
from hashlib import sha256


def execute_attrib_txn(looper, sdk_pool_handle, sdk_wallet_steward, xhash, raw, enc):
    _, identifier = sdk_wallet_steward
    request = looper.loop.run_until_complete(build_attrib_request(identifier, identifier, xhash, raw, enc))
    return sdk_get_reply(looper, sdk_sign_and_submit_req(sdk_pool_handle, sdk_wallet_steward, request))[1]


def test_attrib_xhash_reply_is_valid(looper, sdk_pool_handle, sdk_wallet_steward):
    xhash = sha256("Hello, world".encode()).hexdigest()
    reply = execute_attrib_txn(looper, sdk_pool_handle, sdk_wallet_steward, xhash, None, None)

    validate_write_reply(reply)
    validate_attrib_txn(reply['result']['txn'])
    assert reply['result']['txn']['data']['hash'] == xhash


def test_attrib_raw_reply_is_valid(looper, sdk_pool_handle, sdk_wallet_steward):
    raw = json.dumps({'answer': 42})
    reply = execute_attrib_txn(looper, sdk_pool_handle, sdk_wallet_steward, None, raw, None)

    validate_write_reply(reply)
    validate_attrib_txn(reply['result']['txn'])
    assert json.loads(reply['result']['txn']['data']['raw']) == json.loads(raw)


def test_attrib_enc_reply_is_valid(looper, sdk_pool_handle, sdk_wallet_steward):
    enc = "amgine"
    reply = execute_attrib_txn(looper, sdk_pool_handle, sdk_wallet_steward, None, None, enc)

    validate_write_reply(reply)
    validate_attrib_txn(reply['result']['txn'])
    assert reply['result']['txn']['data']['enc'] == enc



import json

import pytest
from plenum.common.constants import TRUSTEE

from indy_common.authorize.auth_actions import ADD_PREFIX, EDIT_PREFIX
from indy_common.authorize.auth_constraints import AuthConstraint, IDENTITY_OWNER
from indy_common.constants import ATTRIB
from indy_node.test.auth_rule.auth_framework.basic import AuthTest
from indy_node.test.helper import (
    sdk_add_attribute_and_check, build_auth_rule_request_json
)
from plenum.common.exceptions import RequestRejectedException
from plenum.test.pool_transactions.helper import sdk_add_new_nym


class EditAttribTest(AuthTest):
    def __init__(self, env, action_id):
        super().__init__(env, action_id)
        self.attr_data = None

    def prepare(self):
        self.default_auth_rule = self.get_default_auth_rule()
        self.changed_auth_rule = self.get_changed_auth_rule()

        self.attr_data = {'name': 'Drogon'}

    def run(self):
        # Step 1. Check default auth rule
        sdk_add_attribute_and_check(self.looper, self.sdk_pool_handle, self.new_default_wallet,
                                    json.dumps(self.attr_data), self.new_default_wallet[1])
        self.attr_data['name'] = 'Robert'
        with pytest.raises(RequestRejectedException):
            sdk_add_attribute_and_check(self.looper, self.sdk_pool_handle, self.trustee_wallet,
                                        json.dumps(self.attr_data), self.new_default_wallet[1])

        # Step 2. Change auth rule
        self.send_and_check(self.changed_auth_rule, wallet=self.trustee_wallet)

        # Step 3. Check, that we cannot edit attrib the old way
        self.attr_data['name'] = 'Gendry'
        with pytest.raises(RequestRejectedException):
            sdk_add_attribute_and_check(self.looper, self.sdk_pool_handle, self.new_default_wallet,
                                        json.dumps(self.attr_data), self.new_default_wallet[1])

        # Step 4. Check, that new auth rule is used
        self.attr_data['name'] = 'Jaime'
        sdk_add_attribute_and_check(self.looper, self.sdk_pool_handle, self.trustee_wallet,
                                    json.dumps(self.attr_data), self.new_default_wallet[1])

        # Step 5. Return default auth rule
        self.send_and_check(self.default_auth_rule, wallet=self.trustee_wallet)

        # Step 6. Check, that default auth rule works
        self.attr_data['name'] = 'Eddard'
        sdk_add_attribute_and_check(self.looper, self.sdk_pool_handle, self.new_default_wallet,
                                    json.dumps(self.attr_data), self.new_default_wallet[1])
        self.attr_data['name'] = 'Cersei'
        with pytest.raises(RequestRejectedException):
            sdk_add_attribute_and_check(self.looper, self.sdk_pool_handle, self.trustee_wallet,
                                        json.dumps(self.attr_data), self.new_default_wallet[1])

    def result(self):
        pass

    def get_changed_auth_rule(self):
        self.new_default_wallet = sdk_add_new_nym(self.looper, self.sdk_pool_handle, self.trustee_wallet,
                                                  role=IDENTITY_OWNER)
        constraint = AuthConstraint(role=TRUSTEE,
                                    sig_count=1,
                                    need_to_be_owner=False)
        return build_auth_rule_request_json(
            self.looper, self.trustee_wallet[1],
            auth_action=EDIT_PREFIX,
            auth_type=ATTRIB,
            field='*',
            new_value='*',
            old_value='*',
            constraint=constraint.as_dict
        )



from indy_common.authorize.auth_actions import AuthActionAdd, AuthActionEdit
from indy_common.authorize.auth_request_validator import WriteRequestValidator
from plenum.common.constants import TXN_AUTHOR_AGREEMENT_VERSION
from plenum.server.database_manager import DatabaseManager
from plenum.server.request_handlers.static_taa_helper import StaticTAAHelper
from plenum.server.request_handlers.txn_author_agreement_handler import TxnAuthorAgreementHandler \
    as PTxnAuthorAgreementHandler


class TxnAuthorAgreementHandler(PTxnAuthorAgreementHandler):

    def __init__(self, database_manager: DatabaseManager,
                 write_req_validator: WriteRequestValidator):
        super().__init__(database_manager)
        self.write_req_validator = write_req_validator

    def authorize(self, request):
        version = request.operation.get(TXN_AUTHOR_AGREEMENT_VERSION)
        if StaticTAAHelper.get_taa_digest(self.state, version, isCommitted=False) is None:
            self.write_req_validator.validate(request,
                                              [AuthActionAdd(txn_type=self.txn_type,
                                                             field='*',
                                                             value='*')])
        else:
            self.write_req_validator.validate(request,
                                              [AuthActionEdit(txn_type=self.txn_type,
                                                              field='*',
                                                              old_value='*',
                                                              new_value='*')])



from collections import OrderedDict

from indy_common.constants import ROLE, CLAIM_DEF_SIGNATURE_TYPE, CLAIM_DEF_SCHEMA_REF
from plenum.common.constants import TXN_TYPE, TARGET_NYM, \
    DATA, ENC, RAW, HASH, ALIAS, TXN_TIME, VERKEY
from plenum.common.types import f


def getTxnOrderedFields():
    return OrderedDict([
        (f.IDENTIFIER.nm, (str, str)),
        (f.REQ_ID.nm, (str, int)),
        (f.SIG.nm, (str, str)),
        (TXN_TIME, (str, int)),
        (TXN_TYPE, (str, str)),
        (TARGET_NYM, (str, str)),
        (VERKEY, (str, str)),
        (DATA, (str, str)),
        (ALIAS, (str, str)),
        (RAW, (str, str)),
        (ENC, (str, str)),
        (HASH, (str, str)),
        (ROLE, (str, str)),
        (CLAIM_DEF_SCHEMA_REF, (str, str)),
        (CLAIM_DEF_SIGNATURE_TYPE, (str, str))
    ])



import json
import random

import base58
from indy.anoncreds import issuer_create_schema
from indy.ledger import build_schema_request

from indy_common.constants import RS_ID, RS_TYPE, RS_NAME, RS_VERSION, RS_CONTENT
from plenum.common.constants import TXN_TYPE
from plenum.test.helper import sdk_get_reply, sdk_sign_and_submit_req, sdk_get_and_check_replies, sdk_gen_request


def req_id():
    id = random.randint(1, 100000000)
    while True:
        yield id
        id += 1


_reqId = req_id()


# Utility predicates


def is_one_of(*args):
    def check(v):
        return v in args

    return check


def is_int(v):
    return isinstance(v, int)


def is_list(v):
    return isinstance(v, list)


def is_dict(v):
    return isinstance(v, dict)


def is_str(v):
    return isinstance(v, str)


def is_base58_str(v):
    return is_str(v) and all(c in base58.alphabet for c in v.encode())


def is_number_str(v):
    return is_str(v) and v.isdigit()


def is_did(v):
    return is_base58_str(v) and len(base58.b58decode(v)) in [16, 32]


def is_verkey(v):
    if not is_str(v):
        return False
    key_len = 32
    if v.startswith('~'):
        key_len = 16
        v = v[1:]
    return is_base58_str(v) and len(base58.b58decode(v)) == key_len


def is_json(v):
    try:
        json.loads(v)
        return True
    except:
        return False


def is_sha256(v):
    return is_str(v) \
           and all(c in '0123456789abcdef' for c in v) \
           and len(v) == 256 // 4


# Basic requirement checkers

def require(d, k, t):
    assert k in d
    assert t(d[k])


def optional(d, k, t):
    if k in d:
        assert t(d[k])


# Generic validators

def validate_txn(txn):
    require(txn, 'type', is_number_str)
    require(txn, 'data', is_dict)
    require(txn, 'metadata', is_dict)
    optional(txn, 'protocolVersion', is_int)

    metadata = txn['metadata']
    require(metadata, 'from', is_did)
    require(metadata, 'reqId', is_int)


def validate_txn_metadata(txn_metadata):
    require(txn_metadata, 'txnTime', is_int)
    require(txn_metadata, 'seqNo', is_int)
    require(txn_metadata, 'txnId', is_str)


def validate_req_signature(req_signature):
    require(req_signature, 'type', is_one_of('ED25519', 'ED25519_MULTI'))
    require(req_signature, 'values', is_list)
    for value in req_signature['values']:
        assert is_dict(value)
        require(value, 'from', is_did)
        require(value, 'value', is_base58_str)


def validate_write_result(result):
    require(result, 'ver', is_str)
    require(result, 'txn', is_dict)
    require(result, 'txnMetadata', is_dict)
    require(result, 'reqSignature', is_dict)
    require(result, 'rootHash', is_base58_str)
    require(result, 'auditPath', is_list)
    assert all(is_base58_str(v) for v in result['auditPath'])

    validate_txn(result['txn'])
    validate_txn_metadata(result['txnMetadata'])
    validate_req_signature(result['reqSignature'])


def validate_write_reply(reply):
    require(reply, 'op', is_one_of('REPLY'))
    require(reply, 'result', is_dict)
    validate_write_result(reply['result'])


# Specific txn validators

def validate_nym_txn(txn):
    require(txn, 'type', is_one_of('1'))

    data = txn['data']
    require(data, 'dest', is_did)
    optional(data, 'role', is_one_of(None, '0', '2', '101'))
    optional(data, 'verkey', is_verkey)
    optional(data, 'alias', is_str)


def validate_attrib_txn(txn):
    require(txn, 'type', is_one_of('100'))

    data = txn['data']
    require(data, 'dest', is_did)
    optional(data, 'raw', is_json)
    optional(data, 'hash', is_sha256)
    optional(data, 'enc', is_str)
    assert sum(1 for k in data.keys() if k in ['raw', 'hash', 'enc']) == 1


def validate_schema_txn(txn):
    require(txn, 'type', is_one_of('101'))

    data = txn['data']
    require(data, 'data', is_dict)

    data = data['data']
    require(data, 'name', is_str)
    require(data, 'version', is_str)
    require(data, 'attr_names', is_list)
    assert all(is_str(n) for n in data['attr_names'])


def validate_claim_def_txn(txn):
    require(txn, 'type', is_one_of('102'))

    data = txn['data']
    require(data, 'data', is_dict)
    require(data['data'], 'primary', is_dict)
    require(data['data'], 'revocation', is_dict)
    require(data, 'ref', is_int)
    require(data, 'signature_type', is_one_of('CL'))
    optional(data, 'tag', is_str)


def validate_rich_schema_txn(txn, txn_type):
    require(txn, 'type', is_one_of(txn_type))

    data = txn['data']
    require(data, 'id', is_str)
    require(data, 'rsName', is_str)
    require(data, 'rsType', is_str)
    require(data, 'rsVersion', is_str)
    require(data, 'content', is_str)


# Misc utility


def sdk_build_schema_request(looper, sdk_wallet_client,
                             attributes=[], name="", version=""):
    _, identifier = sdk_wallet_client

    _, schema_json = looper.loop.run_until_complete(
        issuer_create_schema(
            identifier, name,
            version, json.dumps(attributes)
        ))

    return looper.loop.run_until_complete(
        build_schema_request(identifier, schema_json)
    )


def build_get_rs_schema_request(did, txnId):
    identifier, type, name, version = txnId.split(':')
    # _id = identifier + ':' + type + ':' + name + ':' + version
    txn_dict = {
        'operation': {
            'type': "301",
            'from': identifier,
            'meta': {
                'name': name,
                'version': version,
                'type': 'sch'  # type
            }
        },
        "identifier": did,
        "reqId": next(_reqId),
        "protocolVersion": 2
    }
    schema_json = json.dumps(txn_dict)
    return schema_json


def build_rs_schema_request(identifier, schema={}, name="", version=""):
    txn_dict = {
        'operation': {
            'type': "201",
            'meta': {
                'name': name,
                'version': version,
                'type': "sch"
            },
            'data': {
                'schema': schema
            }
        },
        "identifier": identifier,
        "reqId": next(_reqId),
        "protocolVersion": 2
    }
    schema_json = json.dumps(txn_dict)
    return schema_json


def sdk_write_schema(looper, sdk_pool_handle, sdk_wallet_client, multi_attribute=[], name="", version=""):
    _, identifier = sdk_wallet_client
    if multi_attribute:
        _, schema_json = looper.loop.run_until_complete(
            issuer_create_schema(identifier, name, version, json.dumps(multi_attribute)))
    else:
        _, schema_json = looper.loop.run_until_complete(
            issuer_create_schema(identifier, "name", "1.0", json.dumps(["first", "last"])))
    request = looper.loop.run_until_complete(build_schema_request(identifier, schema_json))
    return schema_json, \
           sdk_get_reply(looper, sdk_sign_and_submit_req(sdk_pool_handle, sdk_wallet_client, request))[1]


def sdk_write_schema_and_check(looper, sdk_pool_handle, sdk_wallet_client,
                               attributes=[], name="", version=""):
    request = sdk_build_schema_request(looper, sdk_wallet_client,
                                       attributes, name, version)
    req = sdk_sign_and_submit_req(sdk_pool_handle, sdk_wallet_client, request)
    rep = sdk_get_and_check_replies(looper, [req])
    return rep


# Rich Schema


def sdk_build_rich_schema_request(looper, sdk_wallet_client,
                                  txn_type, rs_id, rs_name, rs_version, rs_type, rs_content):
    # TODO: replace by real SDK call
    _, identifier = sdk_wallet_client
    op = {
        TXN_TYPE: txn_type,
        RS_ID: rs_id,
        RS_NAME: rs_name,
        RS_TYPE: rs_type,
        RS_VERSION: rs_version,
        RS_CONTENT: rs_content
    }
    req_obj = sdk_gen_request(op, identifier=sdk_wallet_client[1])
    return json.dumps(req_obj.as_dict)


def sdk_write_rich_schema_object_and_check(looper, sdk_wallet_client, sdk_pool_handle,
                                           txn_type, rs_id, rs_name, rs_version, rs_type, rs_content):
    request = sdk_build_rich_schema_request(looper, sdk_wallet_client,
                                            txn_type, rs_id=rs_id, rs_name=rs_name,
                                            rs_version=rs_version, rs_type=rs_type,
                                            rs_content=json.dumps(rs_content))
    req = sdk_sign_and_submit_req(sdk_pool_handle, sdk_wallet_client, request)
    rep = sdk_get_and_check_replies(looper, [req])
    return rep


def sdk_write_rich_schema_object(looper, sdk_wallet_client, sdk_pool_handle,
                                 txn_type, rs_id, rs_name, rs_version, rs_type, rs_content):
    request = sdk_build_rich_schema_request(looper, sdk_wallet_client,
                                            txn_type, rs_id=rs_id, rs_name=rs_name,
                                            rs_version=rs_version, rs_type=rs_type,
                                            rs_content=json.dumps(rs_content))

    return sdk_get_reply(looper,
                         sdk_sign_and_submit_req(sdk_pool_handle, sdk_wallet_client, req))[1]



from typing import Optional

from indy_node.server.pool_config import PoolConfig

from indy_common.authorize.auth_actions import AuthActionEdit
from indy_common.authorize.auth_request_validator import WriteRequestValidator
from indy_common.constants import POOL_CONFIG, CONFIG_LEDGER_ID, ACTION
from plenum.common.request import Request
from plenum.server.database_manager import DatabaseManager
from plenum.server.request_handlers.handler_interfaces.write_request_handler import WriteRequestHandler


class PoolConfigHandler(WriteRequestHandler):

    def __init__(self, database_manager: DatabaseManager,
                 write_req_validator: WriteRequestValidator,
                 pool_config: PoolConfig):
        super().__init__(database_manager, POOL_CONFIG, CONFIG_LEDGER_ID)
        self.write_req_validator = write_req_validator
        self.pool_config = pool_config

    def static_validation(self, request: Request):
        self._validate_request_type(request)

    def additional_dynamic_validation(self, request: Request, req_pp_time: Optional[int]):
        self._validate_request_type(request)
        action = '*'
        status = '*'
        self.write_req_validator.validate(request,
                                          [AuthActionEdit(txn_type=self.txn_type,
                                                          field=ACTION,
                                                          old_value=status,
                                                          new_value=action)])

    def apply_forced_request(self, req: Request):
        super().apply_forced_request(req)
        txn = self._req_to_txn(req)
        self.pool_config.handleConfigTxn(txn)

    # Config handler don't use state for any validation for now
    def update_state(self, txn, prev_result, request, is_committed=False):
        pass



import pytest

from indy_common.constants import CRED_DEF_ID, REVOC_TYPE, \
    TAG
from indy_node.server.request_handlers.domain_req_handlers.revoc_reg_def_handler import RevocRegDefHandler
from indy_node.test.request_handlers.helper import add_to_idr
from plenum.common.constants import TRUSTEE
from plenum.common.exceptions import InvalidClientRequest, UnauthorizedClientRequest
from plenum.common.txn_util import reqToTxn, append_txn_metadata, get_payload_data
from plenum.server.request_handlers.utils import encode_state_value


def test_revoc_reg_def_dynamic_validation_fails_wrong_id(revoc_reg_def_handler,
                                                         revoc_reg_def_request):
    revoc_reg_def_request.operation[CRED_DEF_ID] = 'sample' * 3
    with pytest.raises(InvalidClientRequest,
                       match="Format of {} field is not acceptable. "
                             "Expected: 'did:marker:signature_type:schema_ref' or "
                             "'did:marker:signature_type:schema_ref:tag'".format(CRED_DEF_ID)):
        revoc_reg_def_handler.static_validation(revoc_reg_def_request)


def test_revoc_reg_def_dynamic_validation_fails_no_cred_def(revoc_reg_def_handler,
                                                            revoc_reg_def_request):
    add_to_idr(revoc_reg_def_handler.database_manager.idr_cache,
               revoc_reg_def_request.identifier,
               TRUSTEE)
    operation = revoc_reg_def_request.operation
    cred_def_id = operation.get(CRED_DEF_ID)
    revoc_def_type = operation.get(REVOC_TYPE)
    revoc_def_tag = operation.get(TAG)
    assert cred_def_id
    assert revoc_def_tag
    assert revoc_def_type
    revoc_def_id = RevocRegDefHandler.make_state_path_for_revoc_def(revoc_reg_def_request.identifier,
                                                                    cred_def_id, revoc_def_type,
                                                                    revoc_def_tag)
    revoc_reg_def_handler.state.set(revoc_def_id, "{}")
    with pytest.raises(InvalidClientRequest,
                       match="There is no any CRED_DEF by path"):
        revoc_reg_def_handler.dynamic_validation(revoc_reg_def_request, 0)


def test_revoc_reg_def_dynamic_validation_passes(revoc_reg_def_handler,
                                                 revoc_reg_def_request):
    add_to_idr(revoc_reg_def_handler.database_manager.idr_cache,
               revoc_reg_def_request.identifier,
               TRUSTEE)
    cred_def_id = revoc_reg_def_request.operation.get(CRED_DEF_ID)
    revoc_reg_def_handler.state.set(cred_def_id.encode(),
                                    encode_state_value("value", "seqNo", "txnTime"))
    revoc_reg_def_handler.dynamic_validation(revoc_reg_def_request, 0)


def test_revoc_reg_def_dynamic_validation_without_permission(revoc_reg_def_handler,
                                                             revoc_reg_def_request):
    add_to_idr(revoc_reg_def_handler.database_manager.idr_cache,
               revoc_reg_def_request.identifier,
               None)
    cred_def_id = revoc_reg_def_request.operation.get(CRED_DEF_ID)
    revoc_reg_def_handler.state.set(cred_def_id.encode(),
                                    encode_state_value("value", "seqNo", "txnTime"))
    with pytest.raises(UnauthorizedClientRequest,
                       match="Not enough TRUSTEE signatures"):
        revoc_reg_def_handler.dynamic_validation(revoc_reg_def_request, 0)


def test_update_state(revoc_reg_def_handler, revoc_reg_def_request):
    seq_no = 1
    txn_time = 1560241033
    txn = reqToTxn(revoc_reg_def_request)
    append_txn_metadata(txn, seq_no, txn_time)
    value = get_payload_data(txn)
    path = RevocRegDefHandler.prepare_revoc_def_for_state(txn,
                                                          path_only=True)

    revoc_reg_def_handler.update_state(txn, None, revoc_reg_def_request)
    assert revoc_reg_def_handler.get_from_state(path) == (value, seq_no, txn_time)



#!/usr/bin/python3.5
import grp
import os
import pwd
import shutil
import stat
import sys
import tarfile
import traceback
import copy
from _sha256 import sha256
from binascii import hexlify

from common.serializers.serialization import ledger_txn_serializer, serialize_msg_for_signing
from indy_common.config_helper import NodeConfigHelper
from indy_common.config_util import getConfig
from indy_common.constants import CONFIG_LEDGER_ID, REVOC_REG_DEF, CRED_DEF_ID, \
    CLAIM_DEF_TAG_DEFAULT, NYM, ATTRIB, SCHEMA, CLAIM_DEF, REVOC_REG_ENTRY, REVOC_REG_DEF_ID, REVOC_TYPE, TAG, ID
from indy_common.state import domain
from indy_node.persistence.attribute_store import AttributeStore
from indy_node.server.config_req_handler import ConfigReqHandler
from indy_node.server.domain_req_handler import DomainReqHandler
from indy_node.server.pool_req_handler import PoolRequestHandler
from plenum.common.constants import TXN_TYPE, TXN_PAYLOAD, \
    TXN_PAYLOAD_METADATA, TXN_PAYLOAD_METADATA_DIGEST, POOL_LEDGER_ID, DOMAIN_LEDGER_ID, TARGET_NYM, RAW, \
    TXN_PAYLOAD_DATA, ENC, HASH
from plenum.common.txn_util import transform_to_new_format, get_from, get_req_id, get_payload_data, \
    get_protocol_version, get_seq_no, get_type, append_txn_metadata
from plenum.common.types import f, OPERATION
from plenum.persistence.req_id_to_txn import ReqIdrToTxn
from storage.helper import initKeyValueStorage
from storage.kv_store_rocksdb_int_keys import KeyValueStorageRocksdbIntKeys
from stp_core.common.log import getlogger

logger = getlogger()

ENV_FILE_PATH = "/etc/indy/indy.env"
ledger_types = ['pool', 'domain', 'config']
config = getConfig()


def set_own_perm(usr, dir):
    uid = pwd.getpwnam(usr).pw_uid
    gid = grp.getgrnam(usr).gr_gid
    perm_mask_rw = stat.S_IRUSR | stat.S_IWUSR | stat.S_IRGRP | stat.S_IWGRP
    perm_mask_rwx = stat.S_IRUSR | stat.S_IWUSR | stat.S_IXUSR | stat.S_IRGRP | stat.S_IWGRP | stat.S_IXGRP

    os.chown(dir, uid, gid)
    os.chmod(dir, perm_mask_rwx)
    for croot, sub_dirs, cfiles in os.walk(dir):
        for fs_name in sub_dirs:
            os.chown(os.path.join(croot, fs_name), uid, gid)
            os.chmod(os.path.join(croot, fs_name), perm_mask_rwx)
        for fs_name in cfiles:
            os.chown(os.path.join(croot, fs_name), uid, gid)
            os.chmod(os.path.join(croot, fs_name), perm_mask_rw)


def get_node_name():
    node_name = None
    node_name_key = 'NODE_NAME'

    if os.path.exists(ENV_FILE_PATH):
        with open(ENV_FILE_PATH, "r") as fenv:
            for line in fenv.readlines():
                if line.find(node_name_key) != -1:
                    node_name = line.split('=')[1].strip()
                    break
    else:
        logger.error("Path to env file does not exist")

    return node_name


def get_ledger_id_by_txn_type(txn_type):

    def get_types_for_req_handler(req_handler):
        return list(req_handler.write_types) + list(req_handler.query_types)

    if txn_type in get_types_for_req_handler(PoolRequestHandler):
        return POOL_LEDGER_ID
    elif txn_type in get_types_for_req_handler(DomainReqHandler):
        return DOMAIN_LEDGER_ID
    elif txn_type in get_types_for_req_handler(ConfigReqHandler):
        return CONFIG_LEDGER_ID
    else:
        logger.error("Unknown txn_type: {}".format(txn_type))
        logger.error("Cannot write txn into SeqNoDB, because cannot define ledger_id")
        sys.exit(1)


def add_tag_into_cred_def_id(val):
    new_val = copy.deepcopy(val)
    old_cred_def_id = new_val.get(CRED_DEF_ID, None)
    if old_cred_def_id:
        path_elems = old_cred_def_id.split(':')
        if len(path_elems) == 4:
            did, marker, sig_type, schema_seq_no = path_elems
            new_cred_def_id = domain.make_state_path_for_claim_def(authors_did=did,
                                                                   schema_seq_no=schema_seq_no,
                                                                   signature_type=sig_type,
                                                                   tag=CLAIM_DEF_TAG_DEFAULT)
            new_val[CRED_DEF_ID] = new_cred_def_id.decode()
            rev_type = new_val.get(REVOC_TYPE)
            rev_tag = new_val.get(TAG)
            if not all([rev_type, rev_tag]):
                return False
            new_revoc_reg_def_id = domain.make_state_path_for_revoc_def(authors_did=did,
                                                                        cred_def_id=new_cred_def_id.decode(),
                                                                        revoc_def_type=rev_type,
                                                                        revoc_def_tag=rev_tag)
            new_val[ID] = new_revoc_reg_def_id.decode()
        else:
            return False
    else:
        return False
    return new_val


def add_cred_def_id_into_entry(val):
    new_val = copy.deepcopy(val)
    old_revoc_reg_def_id = new_val.get(REVOC_REG_DEF_ID, None)
    if old_revoc_reg_def_id:
        path_elems = old_revoc_reg_def_id.split(':')
        did = path_elems[0]
        cred_def_id = ":".join(path_elems[2:-2])
        rev_tag = path_elems[-1]
        rev_type = path_elems[-2]
        new_revoc_reg_def_id = domain.make_state_path_for_revoc_def(authors_did=did,
                                                                    cred_def_id="{}:{}".format(cred_def_id,
                                                                                               CLAIM_DEF_TAG_DEFAULT),
                                                                    revoc_def_type=rev_type,
                                                                    revoc_def_tag=rev_tag)
        new_val[REVOC_REG_DEF_ID] = new_revoc_reg_def_id.decode()
    else:
        return False
    return new_val


def gen_txn_path(txn, attr_store):
    txn_type = get_type(txn)
    if txn_type not in DomainReqHandler.write_types:
        return None

    if txn_type == NYM:
        nym = get_payload_data(txn).get(TARGET_NYM)
        binary_digest = domain.make_state_path_for_nym(nym)
        return hexlify(binary_digest).decode()
    elif txn_type == ATTRIB:
        attr_data = get_payload_data(txn)
        if RAW in attr_data or ENC in attr_data:
            attr_type = RAW if RAW in attr_data else ENC
            attr = attr_store.get(attr_data[attr_type])
            copy_txn = copy.deepcopy(txn)
            copy_txn[TXN_PAYLOAD][TXN_PAYLOAD_DATA][attr_type] = attr
            path = domain.prepare_attr_for_state(copy_txn, path_only=True)
            return path.decode()
        if HASH in attr_data:
            path = domain.prepare_attr_for_state(txn, path_only=True)
            return path.decode()

    elif txn_type == SCHEMA:
        path = domain.prepare_schema_for_state(txn, path_only=True)
        return path.decode()
    elif txn_type == CLAIM_DEF:
        path = domain.prepare_claim_def_for_state(txn, path_only=True)
        return path.decode()
    elif txn_type == REVOC_REG_DEF:
        path = domain.prepare_revoc_def_for_state(txn, path_only=True)
        return path.decode()
    elif txn_type == REVOC_REG_ENTRY:
        path = domain.prepare_revoc_reg_entry_for_state(txn, path_only=True)
        return path.decode()


def migrate_txn_log(db_dir, db_name):

    def put_into_seq_no_db(txn):
        # If there is no reqId, then it's genesis txn
        if get_req_id(txn) is None:
            return
        txn_new = copy.deepcopy(txn)
        operation = get_payload_data(txn_new)
        operation[TXN_TYPE] = get_type(txn_new)
        dct = {
            f.IDENTIFIER.nm: get_from(txn_new),
            f.REQ_ID.nm: get_req_id(txn_new),
            OPERATION: operation,
        }
        if get_protocol_version(txn_new) is not None:
            dct[f.PROTOCOL_VERSION.nm] = get_protocol_version(txn_new)
        digest = sha256(serialize_msg_for_signing(dct)).hexdigest()
        seq_no = get_seq_no(txn_new)
        ledger_id = get_ledger_id_by_txn_type(operation[TXN_TYPE])
        line_to_record = str(ledger_id) + ReqIdrToTxn.delimiter + str(seq_no)
        dest_seq_no_db_storage.put(digest, line_to_record)
        return digest

    new_db_name = db_name + '_new'
    old_path = os.path.join(db_dir, db_name)
    new_path = os.path.join(db_dir, new_db_name)
    new_seqno_db_name = config.seqNoDbName + '_new'
    attr_store = AttributeStore(
        initKeyValueStorage(
            config.attrStorage,
            db_dir,
            config.attrDbName,
            db_config=config.db_attr_db_config))
    try:
        dest_seq_no_db_storage = initKeyValueStorage(config.reqIdToTxnStorage,
                                                     db_dir,
                                                     new_seqno_db_name)
    except Exception:
        logger.error(traceback.format_exc())
        logger.error("Could not open new seq_no_db storage")
        return False

    # open new and old ledgers
    try:
        src_storage = KeyValueStorageRocksdbIntKeys(db_dir, db_name, read_only=True)
    except Exception:
        logger.error(traceback.format_exc())
        logger.error("Could not open old ledger: {}".format(os.path.join(db_dir, db_name)))
        return False

    try:
        dest_storage = KeyValueStorageRocksdbIntKeys(db_dir, new_db_name)
    except Exception:
        logger.error(traceback.format_exc())
        logger.error("Could not open new ledger: {}".format(os.path.join(db_dir, new_db_name)))
        return False

    # put values from old ledger to the new one
    try:
        for key, val in src_storage.iterator():
            key = key.decode()
            val = ledger_txn_serializer.deserialize(val)
            if val.get(TXN_TYPE) == REVOC_REG_DEF:
                val = add_tag_into_cred_def_id(val)
                if val is False:
                    return False
            if val.get(TXN_TYPE) == REVOC_REG_ENTRY:
                val = add_cred_def_id_into_entry(val)
                if val is False:
                    return False
            new_val = transform_to_new_format(txn=val, seq_no=int(key))
            digest = put_into_seq_no_db(new_val)
            # add digest into txn
            if get_req_id(new_val):
                new_val[TXN_PAYLOAD][TXN_PAYLOAD_METADATA][TXN_PAYLOAD_METADATA_DIGEST] = digest
                txn_id = gen_txn_path(new_val, attr_store=attr_store)
                if txn_id:
                    append_txn_metadata(new_val,
                                        txn_id=txn_id)
            new_val = ledger_txn_serializer.serialize(new_val)
            dest_storage.put(key, new_val)

    except Exception:
        logger.error(traceback.format_exc())
        logger.error("Could not put key/value to the new ledger '{}'".format(db_name))
        return False

    src_storage.close()
    dest_storage.close()
    dest_seq_no_db_storage.close()
    attr_store.close()

    # Remove old ledger
    try:
        shutil.rmtree(old_path)
    except Exception:
        logger.error(traceback.format_exc())
        logger.error("Could not remove old ledger: {}"
                     .format(old_path))
        return False

    # Rename new ledger to old one
    try:
        shutil.move(new_path, old_path)
    except Exception:
        logger.error(traceback.format_exc())
        logger.error("Could not rename temporary new ledger from '{}' to '{}'"
                     .format(new_path, old_path))
        return False
    try:
        set_own_perm("indy", old_path)
    except Exception:
        pass

    return True


def rename_seq_no_db(db_dir):
    old_seqno_path = os.path.join(db_dir, config.seqNoDbName)
    new_seqno_db_name = config.seqNoDbName + '_new'
    new_seqno_path = os.path.join(db_dir, new_seqno_db_name)
    try:
        shutil.rmtree(old_seqno_path)
    except Exception:
        logger.error(traceback.format_exc())
        logger.error("Could not remove old seq_no_db: {}"
                     .format(old_seqno_path))
        return False
    try:
        shutil.move(new_seqno_path, old_seqno_path)
    except Exception:
        logger.error(traceback.format_exc())
        logger.error("Could not rename temporary new seq_no_db from '{}' to '{}'"
                     .format(new_seqno_path, old_seqno_path))
        return False

    set_own_perm("indy", old_seqno_path)


def migrate_txn_logs(ledger_dir):
    # Migrate transaction logs, they use integer keys
    for ledger_type in ledger_types:
        db_name = ledger_type + "_transactions"
        if not migrate_txn_log(ledger_dir, db_name):
            logger.error("Could not migrate {}, DB path: {}"
                         .format(db_name, os.path.join(ledger_dir, db_name)))
            return False

    return True


def migrate_hash_stores(ledger_dir):
    for ledger_type in ledger_types:
        leaves_db = os.path.join(ledger_dir, ledger_type + '_merkleLeaves')
        nodes_db = os.path.join(ledger_dir, ledger_type + '_merkleNodes')
        if os.path.exists(leaves_db):
            shutil.rmtree(leaves_db)
        if os.path.exists(nodes_db):
            shutil.rmtree(nodes_db)
    return True


def migrate_states(ledger_dir):
    # just remove, it will be re-created from txn log
    for ledger_type in ledger_types:
        db = os.path.join(ledger_dir, ledger_type + '_state')
        if os.path.exists(db):
            shutil.rmtree(db)
    return True


def migrate_ts_store(ledger_dir):
    # just remove, since state root hash may be changed
    for ledger_type in ledger_types:
        db = os.path.join(ledger_dir, config.stateTsDbName)
        if os.path.exists(db):
            shutil.rmtree(db)
    return True


def migrate_bls_signature_store(ledger_dir):
    # just remove, since state root hash may be changed
    for ledger_type in ledger_types:
        db = os.path.join(ledger_dir, config.stateSignatureDbName)
        if os.path.exists(db):
            shutil.rmtree(db)
    return True


def archive_old_ledger(node_name, ledger_dir):
    ledger_archive_name = node_name + "_old_txn_ledger.tar.gz"
    ledger_archive_path = os.path.join("/tmp", ledger_archive_name)
    tar = tarfile.open(ledger_archive_path, "w:gz")
    tar.add(ledger_dir, arcname=node_name)
    tar.close()
    logger.info("Archive of old transaction format ledger created: {}"
                .format(ledger_archive_path))


def migrate_all():
    node_name = get_node_name()
    if node_name is None:
        logger.error("Could not get node name")
        return False

    config_helper = NodeConfigHelper(node_name, config)

    ledger_dir = config_helper.ledger_dir

    # 1. Archiving old ledger
    try:
        archive_old_ledger(node_name, ledger_dir)
    except Exception:
        logger.warning("Could not create an archive of old transactions ledger, proceed anyway")

    # 2. migrate txn log
    if migrate_txn_logs(ledger_dir):
        logger.info("All txn logs migrated successfully from old to new transaction format")
    else:
        logger.error("Txn log migration from old to new format failed!")
        return False

    # Rename new seq_no_db into old
    rename_seq_no_db(ledger_dir)

    # 3. migrate hash store
    if migrate_hash_stores(ledger_dir):
        logger.info("All hash stores migrated successfully from old to new transaction format")
    else:
        logger.error("Hash store migration from old to new format failed!")
        return False

    # 4. migrate states
    if migrate_states(ledger_dir):
        logger.info("All states migrated successfully from old to new transaction format")
    else:
        logger.error("State migration from old to new format failed!")
        return False

    # 5. migrate ts store
    if migrate_ts_store(ledger_dir):
        logger.info("Timestamp store migrated successfully from old to new transaction format")
    else:
        logger.error("Timestamp store migration from old to new format failed!")
        return False

    # 6. migrate bls signature xtore
    if migrate_bls_signature_store(ledger_dir):
        logger.info("BLS signature store migrated successfully from old to new transaction format")
    else:
        logger.error("BLS signature store migration from old to new format failed!")
        return False

    return True


if migrate_all():
    logger.info("Migration of txns format")
else:
    logger.info("Migration of txns format failed!")
    sys.exit(1)



import json
import random
from abc import ABCMeta, abstractmethod

from indy_common.authorize.auth_actions import EDIT_PREFIX, split_action_id
from indy_common.authorize.auth_map import auth_map
from indy_common.constants import ENDORSER, NETWORK_MONITOR, NETWORK_MONITOR_STRING, ENDORSER_STRING
from plenum.common.constants import TRUSTEE, TRUSTEE_STRING, STEWARD_STRING, STEWARD, IDENTITY_OWNER, \
    IDENTITY_OWNER_STRING, VALIDATOR
from plenum.common.util import randomString
from plenum.test.helper import sdk_json_to_request_object
from plenum.test.pool_transactions.helper import (
    prepare_nym_request, prepare_new_node_data, prepare_node_request
)

from indy_node.test.helper import (
    build_auth_rule_request_json, sdk_send_and_check_req_json
)


roles_to_string = {
    TRUSTEE: TRUSTEE_STRING,
    STEWARD: STEWARD_STRING,
    ENDORSER: ENDORSER_STRING,
    NETWORK_MONITOR: NETWORK_MONITOR_STRING,
    IDENTITY_OWNER: '',
    '': '',
}


class AbstractTest(metaclass=ABCMeta):
    action_id = ""

    @abstractmethod
    def prepare(self):
        pass

    @abstractmethod
    def run(self):
        pass

    @abstractmethod
    def result(self):
        pass

    @abstractmethod
    def down(self):
        pass


class AuthTest(AbstractTest):

    def __init__(self, env, action_id):
        self.looper = env.looper
        self.action_id = action_id
        self.action = split_action_id(action_id)
        self.sdk_pool_handle = env.sdk_pool_handle
        self.trustee_wallet = env.sdk_wallet_trustee
        self.env = env

        self.default_auth_rule = None
        self.changed_auth_rule = None
        self.new_default_wallet = None

    def down(self):
        pass

    def _build_nym(self, creator_wallet, role_string, did, verkey=None, skipverkey=True):
        seed = randomString(32)
        alias = randomString(5)
        nym_request, _ = self.looper.loop.run_until_complete(
            prepare_nym_request(creator_wallet,
                                seed,
                                alias,
                                role_string,
                                dest=did,
                                verkey=verkey,
                                skipverkey=skipverkey))
        return nym_request

    def _build_node(self, steward_wallet_handle, tconf, tdir, services=[VALIDATOR],
                    node_name=None, node_data=None):
        if not node_name:
            node_name = randomString(10)
            print(node_name)
        sigseed, verkey, bls_key, nodeIp, nodePort, clientIp, clientPort, key_proof = \
            prepare_new_node_data(tconf, tdir, node_name) if not node_data else node_data

        # filling node request
        _, steward_did = steward_wallet_handle
        node_request = self.looper.loop.run_until_complete(
            prepare_node_request(steward_did,
                                 new_node_name=node_name,
                                 clientIp=clientIp,
                                 clientPort=clientPort,
                                 nodeIp=nodeIp,
                                 nodePort=nodePort,
                                 bls_key=bls_key,
                                 sigseed=sigseed,
                                 services=services,
                                 key_proof=key_proof))
        node_data = sigseed, verkey, bls_key, nodeIp, nodePort, clientIp, clientPort, key_proof
        return node_request, node_data, node_name

    def get_default_auth_rule(self):
        constraint = auth_map.get(self.action_id)
        return build_auth_rule_request_json(
            self.looper, self.trustee_wallet[1],
            auth_action=self.action.prefix,
            auth_type=self.action.txn_type,
            field=self.action.field,
            old_value=self.action.old_value if self.action.prefix == EDIT_PREFIX else None,
            new_value=self.action.new_value,
            constraint=constraint.as_dict
        )

    def send_and_check(self, req_json, wallet):
        return sdk_send_and_check_req_json(
            self.looper, self.sdk_pool_handle, wallet, req_json
        )[0]



from abc import abstractmethod


class GeneratesRequest:
    @abstractmethod
    def _op(self):
        pass

    @abstractmethod
    def ledgerRequest(self):
        """
        Generates a Request object to be submitted to the ledger.
        :return: a Request to be submitted, or None if it shouldn't be written
        """



from copy import deepcopy
from hashlib import sha256

from common.serializers.serialization import domain_state_serializer
from indy_common.constants import ATTRIB, GET_ATTR, SIGNATURE_TYPE, REVOC_TYPE, TAG, CRED_DEF_ID, REVOC_REG_DEF_ID, \
    CLAIM_DEF_SCHEMA_REF, CLAIM_DEF_PUBLIC_KEYS, SCHEMA_ATTR_NAMES, CLAIM_DEF_FROM, CLAIM_DEF_TAG, \
    CLAIM_DEF_TAG_DEFAULT, CLAIM_DEF_CL
from indy_common.req_utils import get_txn_schema_name, get_txn_claim_def_schema_ref, \
    get_txn_claim_def_public_keys, get_txn_claim_def_signature_type, get_txn_claim_def_tag, get_txn_schema_version, \
    get_txn_schema_attr_names, get_reply_schema_from, get_reply_schema_name, get_reply_schema_version, \
    get_reply_schema_attr_names
from indy_common.serialization import attrib_raw_data_serializer
from indy_common.state.state_constants import MARKER_CONTEXT
from plenum.common.constants import RAW, ENC, HASH, TXN_TIME, \
    TARGET_NYM, DATA, TYPE
from plenum.common.txn_util import get_type, get_payload_data, get_seq_no, get_txn_time, get_from
from plenum.common.types import f

MARKER_ATTR = "1"
MARKER_SCHEMA = "2"
MARKER_CLAIM_DEF = "3"
# TODO: change previous markers in "request refactoring" sprint
MARKER_REVOC_DEF = "4"
MARKER_REVOC_REG_ENTRY = "5"
MARKER_REVOC_REG_ENTRY_ACCUM = "6"
LAST_SEQ_NO = "lsn"
VALUE = "val"
LAST_UPDATE_TIME = "lut"

ALL_ATR_KEYS = [RAW, ENC, HASH]


def make_state_path_for_nym(did) -> bytes:
    # TODO: This is duplicated in plenum.DimainRequestHandler
    return sha256(did.encode()).digest()


def make_state_path_for_attr(did, attr_name, attr_is_hash=False) -> bytes:
    nameHash = sha256(attr_name.encode()).hexdigest() if not attr_is_hash else attr_name
    return "{DID}:{MARKER}:{ATTR_NAME}" \
        .format(DID=did,
                MARKER=MARKER_ATTR,
                ATTR_NAME=nameHash).encode()


def make_state_path_for_context(authors_did, context_name, context_version) -> bytes:
    return "{DID}:{MARKER}:{CONTEXT_NAME}:{CONTEXT_VERSION}" \
        .format(DID=authors_did,
                MARKER=MARKER_CONTEXT,
                CONTEXT_NAME=context_name,
                CONTEXT_VERSION=context_version).encode()


def make_state_path_for_schema(authors_did, schema_name, schema_version) -> bytes:
    return "{DID}:{MARKER}:{SCHEMA_NAME}:{SCHEMA_VERSION}" \
        .format(DID=authors_did,
                MARKER=MARKER_SCHEMA,
                SCHEMA_NAME=schema_name,
                SCHEMA_VERSION=schema_version).encode()


def make_state_path_for_claim_def(authors_did, schema_seq_no, signature_type, tag) -> bytes:
    return "{DID}:{MARKER}:{SIGNATURE_TYPE}:{SCHEMA_SEQ_NO}:{TAG}" \
        .format(DID=authors_did,
                MARKER=MARKER_CLAIM_DEF,
                SIGNATURE_TYPE=signature_type,
                SCHEMA_SEQ_NO=schema_seq_no,
                TAG=tag).encode()


def make_state_path_for_revoc_def(authors_did, cred_def_id, revoc_def_type, revoc_def_tag) -> bytes:
    return "{DID}:{MARKER}:{CRED_DEF_ID}:{REVOC_DEF_TYPE}:{REVOC_DEF_TAG}" \
        .format(DID=authors_did,
                MARKER=MARKER_REVOC_DEF,
                CRED_DEF_ID=cred_def_id,
                REVOC_DEF_TYPE=revoc_def_type,
                REVOC_DEF_TAG=revoc_def_tag).encode()


def make_state_path_for_revoc_reg_entry(revoc_reg_def_id) -> bytes:
    return "{MARKER}:{REVOC_REG_DEF_ID}" \
        .format(MARKER=MARKER_REVOC_REG_ENTRY,
                REVOC_REG_DEF_ID=revoc_reg_def_id).encode()


def make_state_path_for_revoc_reg_entry_accum(revoc_reg_def_id) -> bytes:
    return "{MARKER}:{REVOC_REG_DEF_ID}" \
        .format(MARKER=MARKER_REVOC_REG_ENTRY_ACCUM,
                REVOC_REG_DEF_ID=revoc_reg_def_id).encode()


def prepare_get_nym_for_state(reply):
    data = reply.get(DATA)
    value = None
    if data is not None:
        parsed = domain_state_serializer.deserialize(data)
        parsed.pop(TARGET_NYM, None)
        value = domain_state_serializer.serialize(parsed)
    nym = reply[TARGET_NYM]
    key = make_state_path_for_nym(nym)
    return key, value


def prepare_attr_for_state(txn, path_only=False):
    """
    Make key(path)-value pair for state from ATTRIB or GET_ATTR
    :return: state path, state value, value for attribute store
    """
    assert get_type(txn) == ATTRIB
    txn_data = get_payload_data(txn)
    nym = txn_data[TARGET_NYM]
    attr_type, attr_key, value = parse_attr_txn(txn_data)
    path = make_state_path_for_attr(nym, attr_key, attr_type == HASH)
    if path_only:
        return path
    hashed_value = hash_of(value) if value else ''
    seq_no = get_seq_no(txn)
    txn_time = get_txn_time(txn)
    value_bytes = encode_state_value(hashed_value, seq_no, txn_time)
    return attr_type, path, value, hashed_value, value_bytes


def prepare_claim_def_for_state(txn, path_only=False):
    origin = get_from(txn)
    schema_seq_no = get_txn_claim_def_schema_ref(txn)
    if schema_seq_no is None:
        raise ValueError("'{}' field is absent, "
                         "but it must contain schema seq no".format(CLAIM_DEF_SCHEMA_REF))
    data = get_txn_claim_def_public_keys(txn)
    if data is None:
        raise ValueError("'{}' field is absent, "
                         "but it must contain components of keys"
                         .format(CLAIM_DEF_PUBLIC_KEYS))
    signature_type = get_txn_claim_def_signature_type(txn)
    tag = get_txn_claim_def_tag(txn)
    path = make_state_path_for_claim_def(origin, schema_seq_no, signature_type, tag)
    if path_only:
        return path
    seq_no = get_seq_no(txn)
    txn_time = get_txn_time(txn)
    value_bytes = encode_state_value(data, seq_no, txn_time)
    return path, value_bytes


def prepare_revoc_def_for_state(txn, path_only=False):
    author_did = get_from(txn)
    txn_data = get_payload_data(txn)
    cred_def_id = txn_data.get(CRED_DEF_ID)
    revoc_def_type = txn_data.get(REVOC_TYPE)
    revoc_def_tag = txn_data.get(TAG)
    assert author_did
    assert cred_def_id
    assert revoc_def_type
    assert revoc_def_tag
    path = make_state_path_for_revoc_def(author_did,
                                         cred_def_id,
                                         revoc_def_type,
                                         revoc_def_tag)
    if path_only:
        return path
    seq_no = get_seq_no(txn)
    txn_time = get_txn_time(txn)
    assert seq_no
    assert txn_time
    value_bytes = encode_state_value(txn_data, seq_no, txn_time)
    return path, value_bytes


def prepare_revoc_reg_entry_for_state(txn, path_only=False):
    author_did = get_from(txn)
    txn_data = get_payload_data(txn)
    revoc_reg_def_id = txn_data.get(REVOC_REG_DEF_ID)
    assert author_did
    assert revoc_reg_def_id
    path = make_state_path_for_revoc_reg_entry(revoc_reg_def_id=revoc_reg_def_id)
    if path_only:
        return path

    seq_no = get_seq_no(txn)
    txn_time = get_txn_time(txn)
    assert seq_no
    assert txn_time
    # TODO: do not duplicate seqNo here
    # doing this now just for backward-compatibility
    txn_data = deepcopy(txn_data)
    txn_data[f.SEQ_NO.nm] = seq_no
    txn_data[TXN_TIME] = txn_time
    value_bytes = encode_state_value(txn_data, seq_no, txn_time)
    return path, value_bytes


def prepare_revoc_reg_entry_accum_for_state(txn):
    author_did = get_from(txn)
    txn_data = get_payload_data(txn)
    revoc_reg_def_id = txn_data.get(REVOC_REG_DEF_ID)
    seq_no = get_seq_no(txn)
    txn_time = get_txn_time(txn)
    assert author_did
    assert revoc_reg_def_id
    assert seq_no
    assert txn_time
    path = make_state_path_for_revoc_reg_entry_accum(revoc_reg_def_id=revoc_reg_def_id)

    # TODO: do not duplicate seqNo here
    # doing this now just for backward-compatibility
    txn_data = deepcopy(txn_data)
    txn_data[f.SEQ_NO.nm] = seq_no
    txn_data[TXN_TIME] = txn_time
    value_bytes = encode_state_value(txn_data, seq_no, txn_time)
    return path, value_bytes


def prepare_get_claim_def_for_state(reply):
    origin = reply.get(CLAIM_DEF_FROM)
    schema_seq_no = reply.get(CLAIM_DEF_SCHEMA_REF)
    if schema_seq_no is None:
        raise ValueError("'{}' field is absent, "
                         "but it must contain schema seq no".format(CLAIM_DEF_SCHEMA_REF))

    signature_type = reply.get(SIGNATURE_TYPE, CLAIM_DEF_CL)
    tag = reply.get(CLAIM_DEF_TAG, CLAIM_DEF_TAG_DEFAULT)
    path = make_state_path_for_claim_def(origin, schema_seq_no, signature_type, tag)
    seq_no = reply[f.SEQ_NO.nm]

    value_bytes = None
    data = reply.get(CLAIM_DEF_PUBLIC_KEYS)
    if data is not None:
        txn_time = reply[TXN_TIME]
        value_bytes = encode_state_value(data, seq_no, txn_time)
    return path, value_bytes


def prepare_get_revoc_def_for_state(reply):
    author_did = reply.get(f.IDENTIFIER.nm)
    cred_def_id = reply.get(DATA).get(CRED_DEF_ID)
    revoc_def_type = reply.get(DATA).get(REVOC_TYPE)
    revoc_def_tag = reply.get(DATA).get(TAG)
    assert author_did
    assert cred_def_id
    assert revoc_def_type
    assert revoc_def_tag
    path = make_state_path_for_revoc_def(author_did,
                                         cred_def_id,
                                         revoc_def_type,
                                         revoc_def_tag)
    seq_no = reply[f.SEQ_NO.nm]
    txn_time = reply[TXN_TIME]
    assert seq_no
    assert txn_time
    value_bytes = encode_state_value(reply[DATA], seq_no, txn_time)
    return path, value_bytes


def prepare_get_revoc_reg_entry_for_state(reply):
    revoc_reg_def_id = reply.get(DATA).get(REVOC_REG_DEF_ID)
    assert revoc_reg_def_id
    path = make_state_path_for_revoc_reg_entry(revoc_reg_def_id=revoc_reg_def_id)

    seq_no = reply[f.SEQ_NO.nm]
    txn_time = reply[TXN_TIME]
    assert seq_no
    assert txn_time
    value_bytes = encode_state_value(reply[DATA], seq_no, txn_time)
    return path, value_bytes


def prepare_get_revoc_reg_entry_accum_for_state(reply):
    revoc_reg_def_id = reply.get(DATA).get(REVOC_REG_DEF_ID)
    seq_no = reply[f.SEQ_NO.nm]
    txn_time = reply[TXN_TIME]
    assert revoc_reg_def_id
    assert seq_no
    assert txn_time
    path = make_state_path_for_revoc_reg_entry_accum(revoc_reg_def_id=revoc_reg_def_id)

    value_bytes = encode_state_value(reply[DATA], seq_no, txn_time)
    return path, value_bytes


def prepare_schema_for_state(txn, path_only=False):
    origin = get_from(txn)
    schema_name = get_txn_schema_name(txn)
    schema_version = get_txn_schema_version(txn)
    value = {
        SCHEMA_ATTR_NAMES: get_txn_schema_attr_names(txn)
    }
    path = make_state_path_for_schema(origin, schema_name, schema_version)
    if path_only:
        return path
    seq_no = get_seq_no(txn)
    txn_time = get_txn_time(txn)
    value_bytes = encode_state_value(value, seq_no, txn_time)
    return path, value_bytes


def prepare_get_schema_for_state(reply):
    origin = get_reply_schema_from(reply)
    schema_name = get_reply_schema_name(reply)
    schema_version = get_reply_schema_version(reply)
    path = make_state_path_for_schema(origin, schema_name, schema_version)
    value_bytes = None
    attr_names = get_reply_schema_attr_names(reply)
    if attr_names:
        data = {
            SCHEMA_ATTR_NAMES: attr_names
        }
        seq_no = reply[f.SEQ_NO.nm]
        txn_time = reply[TXN_TIME]
        value_bytes = encode_state_value(data, seq_no, txn_time)
    return path, value_bytes


def encode_state_value(value, seqNo, txnTime):
    return domain_state_serializer.serialize({
        LAST_SEQ_NO: seqNo,
        LAST_UPDATE_TIME: txnTime,
        VALUE: value
    })


def decode_state_value(ecnoded_value):
    decoded = domain_state_serializer.deserialize(ecnoded_value)
    value = decoded.get(VALUE)
    last_seq_no = decoded.get(LAST_SEQ_NO)
    last_update_time = decoded.get(LAST_UPDATE_TIME)
    return value, last_seq_no, last_update_time


def hash_of(text) -> str:
    if not isinstance(text, (str, bytes)):
        text = domain_state_serializer.serialize(text)
    if not isinstance(text, bytes):
        text = text.encode()
    return sha256(text).hexdigest()


def parse_attr_txn(txn_data):
    attr_type, attr = _extract_attr_typed_value(txn_data)

    if attr_type == RAW:
        data = attrib_raw_data_serializer.deserialize(attr)
        # To exclude user-side formatting issues
        re_raw = attrib_raw_data_serializer.serialize(data,
                                                      toBytes=False)
        key, _ = data.popitem()
        return attr_type, key, re_raw
    if attr_type == ENC:
        return attr_type, attr, attr
    if attr_type == HASH:
        return attr_type, attr, None


def prepare_get_attr_for_state(reply):
    nym = reply[TARGET_NYM]
    attr_type, attr_key = _extract_attr_typed_value(reply)
    data = reply.get(DATA)
    if data:
        reply = reply.copy()
        data = reply.pop(DATA)
        reply[attr_type] = data

        assert reply[TYPE] == GET_ATTR
        attr_type, attr_key, value = parse_attr_txn(reply)
        hashed_value = hash_of(value) if value else ''
        seq_no = reply[f.SEQ_NO.nm]
        txn_time = reply[TXN_TIME]
        value_bytes = encode_state_value(hashed_value, seq_no, txn_time)
        path = make_state_path_for_attr(nym, attr_key, attr_type == HASH)

        return attr_type, path, value, hashed_value, value_bytes

    if attr_type == ENC:
        attr_key = hash_of(attr_key)

    path = make_state_path_for_attr(nym, attr_key,
                                    attr_type == HASH or attr_type == ENC)
    return attr_type, path, None, None, None


def _extract_attr_typed_value(txn_data):
    """
    ATTR and GET_ATTR can have one of 'raw', 'enc' and 'hash' fields.
    This method checks which of them presents and return it's name
    and value in it.
    """
    existing_keys = [key for key in ALL_ATR_KEYS if key in txn_data]
    if len(existing_keys) == 0:
        raise ValueError("ATTR should have one of the following fields: {}"
                         .format(ALL_ATR_KEYS))
    if len(existing_keys) > 1:
        raise ValueError("ATTR should have only one of the following fields: {}"
                         .format(ALL_ATR_KEYS))
    existing_key = existing_keys[0]
    return existing_key, txn_data[existing_key]



import pytest

from indy_node.server.request_handlers.read_req_handlers.get_revoc_reg_delta_handler import GetRevocRegDeltaHandler
from indy_node.test.anon_creds.helper import build_get_revoc_reg_delta
from plenum.common.txn_util import reqToTxn, append_txn_metadata, get_txn_time
from plenum.common.types import f
from indy_common.types import Request
from plenum.common.constants import TXN_TIME, DOMAIN_LEDGER_ID
from indy_common.state import domain
from plenum.test.helper import sdk_sign_request_from_dict
from plenum.common.util import randomString
from indy_node.test.anon_creds.conftest import build_revoc_reg_entry_for_given_revoc_reg_def
from indy_common.constants import REVOC_REG_DEF_ID, REVOC_REG_DEF, ISSUANCE_BY_DEFAULT, \
    CRED_DEF_ID, VALUE, TAG, ID, TXN_TYPE, REVOC_TYPE, ISSUANCE_TYPE, MAX_CRED_NUM, \
    TAILS_HASH, TAILS_LOCATION, PUBLIC_KEYS, FROM


FIRST_ID_TS = 1000
SECOND_TS_ID = 2000


@pytest.fixture(scope="module")
def add_another_reg_id(looper,
                       sdk_wallet_steward,
                       create_node_and_not_start):
    node = create_node_and_not_start
    data = {
        ID: randomString(50),
        TXN_TYPE: REVOC_REG_DEF,
        REVOC_TYPE: "CL_ACCUM",
        TAG: randomString(5),
        CRED_DEF_ID: randomString(50),
        VALUE:{
            ISSUANCE_TYPE: ISSUANCE_BY_DEFAULT,
            MAX_CRED_NUM: 1000000,
            TAILS_HASH: randomString(50),
            TAILS_LOCATION: 'http://tails.location.com',
            PUBLIC_KEYS: {},
        }
    }
    req = sdk_sign_request_from_dict(looper, sdk_wallet_steward, data)
    looper.runFor(2)
    txn = append_txn_metadata(reqToTxn(Request(**req)),
                              txn_time=FIRST_ID_TS,
                              seq_no=node.domainLedger.seqNo + 1)
    node.write_manager.update_state(txn)
    return req


@pytest.fixture(scope="function")
def reg_entry_with_other_reg_id(looper,
                                sdk_wallet_steward,
                                add_another_reg_id,
                                create_node_and_not_start):
    node = create_node_and_not_start
    revoc_def_txn = add_another_reg_id
    data = build_revoc_reg_entry_for_given_revoc_reg_def(revoc_def_txn)
    req = sdk_sign_request_from_dict(looper, sdk_wallet_steward, data)
    looper.runFor(2)
    txn = append_txn_metadata(reqToTxn(Request(**req)),
                              txn_time=FIRST_ID_TS,
                              seq_no=node.domainLedger.seqNo + 1)
    node.write_manager.update_state(txn)
    node.db_manager.ts_store.set(get_txn_time(txn), node.db_manager.get_state(DOMAIN_LEDGER_ID).headHash)
    return txn


def test_get_delta_with_other_reg_def_in_state(looper,
                                               create_node_and_not_start,
                                               reg_entry_with_other_reg_id,
                                               build_txn_for_revoc_def_entry_by_default,
                                               sdk_wallet_steward):
    entry_second_id = build_txn_for_revoc_def_entry_by_default
    delta_req = build_get_revoc_reg_delta(looper, sdk_wallet_steward)
    node = create_node_and_not_start
    # need for different txnTime
    looper.runFor(2)
    txn = append_txn_metadata(reqToTxn(entry_second_id),
                              txn_time=SECOND_TS_ID,
                              seq_no=node.domainLedger.seqNo + 1)
    node.write_manager.update_state(txn)
    node.db_manager.ts_store.set(get_txn_time(txn), node.getState(DOMAIN_LEDGER_ID).headHash)

    # timestamp beetween FIRST_ID_TS and SECOND_ID_TS
    delta_req['operation'][FROM] = FIRST_ID_TS + 10
    path_to_reg_entry = domain.make_state_path_for_revoc_reg_entry(
        revoc_reg_def_id=entry_second_id['operation'][REVOC_REG_DEF_ID])

    req_handler = None
    for h in node.read_manager.request_handlers.values():
        if isinstance(h, GetRevocRegDeltaHandler):
            req_handler = h

    reg_entry = req_handler._get_reg_entry_by_timestamp(
        delta_req['operation'][FROM],
        path_to_reg_entry)
    # we found root_hash in txRevoc storage but there is not corresponded reg_entry by path
    assert reg_entry.root_hash is not None
    assert reg_entry.value is None

    path_to_reg_entry_accum = domain.make_state_path_for_revoc_reg_entry(
        revoc_reg_def_id=entry_second_id['operation'][REVOC_REG_DEF_ID])
    reg_entry_accum = req_handler._get_reg_entry_accum_by_timestamp(
        delta_req['operation'][FROM],
        path_to_reg_entry_accum)
    assert reg_entry_accum.root_hash is not None
    assert reg_entry_accum.value is None





import asyncio
import json
import socket
from indy.did import create_and_store_my_did
from indy.error import ErrorCode
from indy.ledger import build_pool_restart_request

from plenum.common.util import randomString
from plenum.test.helper import sdk_get_replies, sdk_check_reply
from plenum.test.pool_transactions.helper import sdk_sign_and_send_prepared_request
from stp_core.common.log import getlogger

from indy_common.constants import MESSAGE_TYPE, \
    RESTART_MESSAGE
from indy_common.config import controlServiceHost, controlServicePort

logger = getlogger()


def compose_restart_message(action):
    return (json.dumps({MESSAGE_TYPE: RESTART_MESSAGE})).encode()


def send_restart_message(action):
    sock = socket.create_connection(
        (controlServiceHost, controlServicePort))
    sock.sendall(compose_restart_message(action))
    sock.close()


async def _createServer(host, port):
    """
    Create async server that listens host:port, reads client request and puts
    value to some future that can be used then for checks

    :return: reference to server and future for request
    """
    indicator = asyncio.Future()

    async def _handle(reader, writer):
        raw = await reader.readline()
        request = raw.decode("utf-8")
        indicator.set_result(request)
    server = await asyncio.start_server(_handle, host, port)
    return server, indicator


def _stopServer(server):
    print('Closing server')
    server.close()
    # def _stop(x):
    #     print('Closing server')
    #     server.close()
    # return _stop


def _checkFuture(future):
    """
    Wrapper for futures that lets checking of their status using 'eventually'
    """
    def _check():
        if future.cancelled():
            return None
        if future.done():
            return future.result()
        raise Exception()
    return _check


def sdk_send_restart(looper, trusty_wallet, sdk_pool_handle,
                     action=None, datetime=None, seed=None):
    # filling restart request and getting trusty did
    seed = seed or randomString(32)
    restart_request, did = looper.loop.run_until_complete(
        prepare_restart_request(trusty_wallet, seed, action, datetime))

    # sending request using 'sdk_' functions
    request_couple = sdk_sign_and_send_prepared_request(looper, trusty_wallet,
                                                        sdk_pool_handle,
                                                        restart_request)
    # waiting for replies
    return sdk_get_and_check_multiply_replies(looper, request_couple)


def sdk_get_and_check_multiply_replies(looper, request_couple):
    rets = []
    for req_res in sdk_get_replies(looper, [request_couple, ]):
        req, responses = req_res
        if not isinstance(responses, ErrorCode) and "op" not in responses:
            for node_resp in responses.values():
                sdk_check_reply((req, json.loads(node_resp)))
        else:
            sdk_check_reply(req_res)
        rets.append(req_res)
    return rets[0]


async def prepare_restart_request(wallet, named_seed, action="start",
                                  restart_time="0"):
    wh, submitter_did = wallet
    (named_did, named_verkey) = \
        await create_and_store_my_did(wh, json.dumps({'seed': named_seed}))
    restart_request = await build_pool_restart_request(submitter_did,
                                                       action,
                                                       restart_time)
    return restart_request, named_did




from indy.ledger import build_nym_request
from indy_node.test.helper import createHalfKeyIdentifierAndAbbrevVerkey
from indy_node.test.api.helper import validate_write_reply, validate_nym_txn
from plenum.test.helper import sdk_get_reply, sdk_sign_and_submit_req


def test_nym_reply_is_valid(looper, sdk_pool_handle, sdk_wallet_steward):
    idr, verkey = createHalfKeyIdentifierAndAbbrevVerkey()

    _, identifier = sdk_wallet_steward
    request = looper.loop.run_until_complete(build_nym_request(identifier, idr, verkey, None, None))
    reply = sdk_get_reply(looper, sdk_sign_and_submit_req(sdk_pool_handle, sdk_wallet_steward, request))[1]

    validate_write_reply(reply)
    validate_nym_txn(reply['result']['txn'])



from _sha256 import sha256
from copy import deepcopy
from json import JSONDecodeError
from typing import Optional

from indy_common.authorize.auth_actions import AuthActionEdit, AuthActionAdd
from indy_common.authorize.auth_request_validator import WriteRequestValidator
from indy_common.serialization import attrib_raw_data_serializer
from indy_common.state import domain

from indy_common.constants import ATTRIB
from indy_common.state.state_constants import MARKER_ATTR
from indy_node.server.request_handlers.utils import validate_attrib_keys
from plenum.common.constants import DOMAIN_LEDGER_ID, RAW, ENC, HASH, TARGET_NYM
from plenum.common.exceptions import InvalidClientRequest

from plenum.common.request import Request
from plenum.common.txn_util import get_type, get_request_data, get_payload_data, get_seq_no, get_txn_time
from plenum.server.database_manager import DatabaseManager
from plenum.server.request_handlers.handler_interfaces.write_request_handler import WriteRequestHandler
from plenum.server.request_handlers.utils import encode_state_value
from stp_core.common.log import getlogger

logger = getlogger()

ALL_ATR_KEYS = [RAW, ENC, HASH]


class AttributeHandler(WriteRequestHandler):

    def __init__(self, database_manager: DatabaseManager,
                 write_req_validator: WriteRequestValidator):
        super().__init__(database_manager, ATTRIB, DOMAIN_LEDGER_ID)
        self.write_req_validator = write_req_validator

    def static_validation(self, request: Request):
        self._validate_request_type(request)
        identifier, req_id, operation = get_request_data(request)

        if not validate_attrib_keys(operation):
            raise InvalidClientRequest(identifier, req_id,
                                       '{} should have one and only one of '
                                       '{}, {}, {}'
                                       .format(ATTRIB, RAW, ENC, HASH))

        if RAW in operation:
            try:
                get_key = attrib_raw_data_serializer.deserialize(operation[RAW])
                if len(get_key) == 0:
                    raise InvalidClientRequest(identifier, request.reqId,
                                               '"raw" attribute field must contain non-empty JSON object {}'.
                                               format(TARGET_NYM))
            except JSONDecodeError:
                raise InvalidClientRequest(identifier, request.reqId,
                                           'Attribute field must be valid JSON object when adding it as a raw field {}'.
                                           format(TARGET_NYM))

    def additional_dynamic_validation(self, request: Request, req_pp_time: Optional[int]):
        self._validate_request_type(request)

        identifier, req_id, operation = get_request_data(request)

        if not (not operation.get(TARGET_NYM) or self.__has_nym(operation[TARGET_NYM], is_committed=False)):
            raise InvalidClientRequest(identifier, request.reqId,
                                       '{} should be added before adding '
                                       'attribute for it'.
                                       format(TARGET_NYM))

        is_owner = self.database_manager.idr_cache.getOwnerFor(operation[TARGET_NYM],
                                                               isCommitted=False) == identifier
        field = None
        value = None
        for key in (RAW, ENC, HASH):
            if key in operation:
                field = key
                value = operation[key]
                break

        if field == RAW:
            get_key = attrib_raw_data_serializer.deserialize(value)
            get_key = list(get_key.keys())[0]
        else:
            get_key = value

        old_value, seq_no, _ = self._get_attr(operation[TARGET_NYM], get_key, field)

        if seq_no is not None:
            self.write_req_validator.validate(request,
                                              [AuthActionEdit(txn_type=ATTRIB,
                                                              field=field,
                                                              old_value=old_value,
                                                              new_value=value,
                                                              is_owner=is_owner)])
        else:
            self.write_req_validator.validate(request,
                                              [AuthActionAdd(txn_type=ATTRIB,
                                                             field=field,
                                                             value=value,
                                                             is_owner=is_owner)])

    def gen_txn_id(self, txn):
        self._validate_txn_type(txn)
        path = self.prepare_attr_for_state(txn, path_only=True)
        return path.decode()

    def update_state(self, txn, prev_result, request, is_committed=False) -> None:
        """
        The state trie stores the hash of the whole attribute data at:
            the did+attribute name if the data is plaintext (RAW)
            the did+hash(attribute) if the data is encrypted (ENC)
        If the attribute is HASH, then nothing is stored in attribute store,
        the trie stores a blank value for the key did+hash
        """
        self._validate_txn_type(txn)
        attr_type, path, value, hashed_value, value_bytes = self.prepare_attr_for_state(txn)
        self.state.set(path, value_bytes)
        if attr_type != HASH:
            self.database_manager.attribute_store.set(hashed_value, value)
        return txn

    def _get_attr(self,
                  did: str,
                  key: str,
                  attr_type,
                  is_commited=False) -> (str, int, int, list):
        assert did is not None
        assert key is not None
        path = AttributeHandler.make_state_path_for_attr(did, key, attr_type == HASH)
        try:
            hashed_val, lastSeqNo, lastUpdateTime = self.get_from_state(path, is_commited)
        except KeyError:
            return None, None, None, None
        if not hashed_val or hashed_val == '':
            # Its a HASH attribute
            return hashed_val, lastSeqNo, lastUpdateTime
        else:
            value = self.database_manager.attribute_store.get(hashed_val)
        return value, lastSeqNo, lastUpdateTime

    def __has_nym(self, nym, is_committed: bool = True):
        return self.database_manager.idr_cache.hasNym(nym, isCommitted=is_committed)

    @staticmethod
    def prepare_attr_for_state(txn, path_only=False):
        """
        Make key(path)-value pair for state from ATTRIB or GET_ATTR
        :return: state path, state value, value for attribute store
        """
        assert get_type(txn) == ATTRIB
        txn_data = get_payload_data(txn)
        nym = txn_data[TARGET_NYM]
        attr_type, attr_key, value = AttributeHandler.parse_attr_txn(txn_data)
        path = AttributeHandler.make_state_path_for_attr(nym, attr_key, attr_type == HASH)
        if path_only:
            return path
        hashed_value = domain.hash_of(value) if value else ''
        seq_no = get_seq_no(txn)
        txn_time = get_txn_time(txn)
        value_bytes = encode_state_value(hashed_value, seq_no, txn_time)
        return attr_type, path, value, hashed_value, value_bytes

    @staticmethod
    def parse_attr_txn(txn_data):
        attr_type, attr = AttributeHandler._extract_attr_typed_value(txn_data)

        if attr_type == RAW:
            data = attrib_raw_data_serializer.deserialize(attr)
            # To exclude user-side formatting issues
            re_raw = attrib_raw_data_serializer.serialize(data,
                                                          toBytes=False)
            key, _ = data.popitem()
            return attr_type, key, re_raw
        if attr_type == ENC:
            return attr_type, attr, attr
        if attr_type == HASH:
            return attr_type, attr, None

    @staticmethod
    def _extract_attr_typed_value(txn_data):
        """
        ATTR and GET_ATTR can have one of 'raw', 'enc' and 'hash' fields.
        This method checks which of them presents and return it's name
        and value in it.
        """
        existing_keys = [key for key in ALL_ATR_KEYS if key in txn_data]
        if len(existing_keys) == 0:
            raise ValueError("ATTR should have one of the following fields: {}"
                             .format(ALL_ATR_KEYS))
        if len(existing_keys) > 1:
            raise ValueError("ATTR should have only one of the following fields: {}"
                             .format(ALL_ATR_KEYS))
        existing_key = existing_keys[0]
        return existing_key, txn_data[existing_key]

    @staticmethod
    def make_state_path_for_attr(did, attr_name, attr_is_hash=False) -> bytes:
        nameHash = sha256(attr_name.encode()).hexdigest() if not attr_is_hash else attr_name
        return "{DID}:{MARKER}:{ATTR_NAME}" \
            .format(DID=did,
                    MARKER=MARKER_ATTR,
                    ATTR_NAME=nameHash).encode()

    def transform_txn_for_ledger(self, txn):
        """
        Creating copy of result so that `RAW`, `ENC` or `HASH` can be
        replaced by their hashes. We do not insert actual attribute data
        in the ledger but only the hash of it.
        """
        txn = deepcopy(txn)
        txn_data = get_payload_data(txn)
        attr_type, _, value = self.parse_attr_txn(txn_data)
        if attr_type in [RAW, ENC]:
            txn_data[attr_type] = domain.hash_of(value) if value else ''

        return txn



import pytest
from indy_common.constants import POOL_UPGRADE, ACTION, START, PACKAGE, APP_NAME, REINSTALL
from indy_node.server.request_handlers.config_req_handlers.pool_upgrade_handler import PoolUpgradeHandler
from plenum.common.constants import VERSION, TXN_PAYLOAD, TXN_PAYLOAD_DATA
from plenum.common.exceptions import InvalidClientRequest

from plenum.common.request import Request
from plenum.common.util import randomString
from plenum.test.testing_utils import FakeSomething

from indy_common.version import src_version_cls
from indy_node.server.upgrader import Upgrader
from indy_node.utils.node_control_utils import NodeControlUtil, DebianVersion


@pytest.fixture(scope='function')
def pool_upgrade_request():
    return Request(identifier=randomString(),
                   reqId=5,
                   operation={
                       'type': POOL_UPGRADE,
                       ACTION: START,
                       PACKAGE: 'smth',
                       VERSION: '1.2.3'
                   })


@pytest.fixture(scope='function')
def pool_upgrade_handler(write_auth_req_validator):
    return PoolUpgradeHandler(
        None,
        FakeSomething(check_upgrade_possible=Upgrader.check_upgrade_possible),
        write_auth_req_validator,
        FakeSomething()
    )


@pytest.fixture(scope='function')
def pkg_version(pool_upgrade_request):
    return DebianVersion(
        '1.1.1',
        upstream_cls=src_version_cls(
            pool_upgrade_request.operation[PACKAGE])
    )


def test_pool_upgrade_static_validation_fails_action(pool_upgrade_handler,
                                                     pool_upgrade_request):
    pool_upgrade_request.operation[ACTION] = 'smth'
    with pytest.raises(InvalidClientRequest) as e:
        pool_upgrade_handler.static_validation(pool_upgrade_request)
    e.match('not a valid action')


def test_pool_upgrade_static_validation_fails_schedule(pool_upgrade_handler,
                                                       pool_upgrade_request):
    pool_upgrade_handler.pool_manager.getNodesServices = lambda: 1
    pool_upgrade_handler.upgrader.isScheduleValid = lambda schedule, node_srvs, force: (False, '')
    with pytest.raises(InvalidClientRequest) as e:
        pool_upgrade_handler.static_validation(pool_upgrade_request)
    e.match('not a valid schedule since')


def test_pool_upgrade_static_validation_passes(pool_upgrade_handler,
                                               pool_upgrade_request):
    pool_upgrade_handler.pool_manager.getNodesServices = lambda: 1
    pool_upgrade_handler.upgrader.isScheduleValid = lambda schedule, node_srvs, force: (True, '')
    pool_upgrade_handler.static_validation(pool_upgrade_request)


def test_pool_upgrade_dynamic_validation_fails_pckg(pool_upgrade_handler,
                                                    pool_upgrade_request,
                                                    tconf):
    pool_upgrade_request.operation[PACKAGE] = ''
    with pytest.raises(InvalidClientRequest) as e:
        pool_upgrade_handler.dynamic_validation(pool_upgrade_request, 0)
    e.match('Upgrade package name is empty')


def test_pool_upgrade_dynamic_validation_fails_not_installed(
        monkeypatch,
        pool_upgrade_handler,
        pool_upgrade_request,
        tconf):
    monkeypatch.setattr(NodeControlUtil, 'curr_pkg_info',
                        lambda *x: (None, None))
    with pytest.raises(InvalidClientRequest) as e:
        pool_upgrade_handler.dynamic_validation(pool_upgrade_request, 0)
    e.match('is not installed and cannot be upgraded')


def test_pool_upgrade_dynamic_validation_fails_belong(
        monkeypatch,
        pool_upgrade_handler,
        pool_upgrade_request,
        tconf):
    monkeypatch.setattr(NodeControlUtil, 'curr_pkg_info',
                        lambda *x: ('1.1.1', ['some_pkg']))
    with pytest.raises(InvalidClientRequest) as e:
        pool_upgrade_handler.dynamic_validation(pool_upgrade_request, 0)
    e.match('doesn\'t belong to pool')


def test_pool_upgrade_dynamic_validation_fails_upgradable(
        monkeypatch,
        pool_upgrade_handler,
        pool_upgrade_request,
        pkg_version,
        tconf):
    monkeypatch.setattr(
        NodeControlUtil, 'curr_pkg_info',
        lambda *x: (pkg_version, [APP_NAME])
    )
    pool_upgrade_request.operation[VERSION] = pkg_version.upstream.full
    pool_upgrade_request.operation[REINSTALL] = False
    with pytest.raises(InvalidClientRequest) as e:
        pool_upgrade_handler.dynamic_validation(pool_upgrade_request, 0)
    e.match('Version {} is not upgradable'.format(pkg_version.upstream.full))


def test_pool_upgrade_dynamic_validation_fails_scheduled(
        monkeypatch,
        pool_upgrade_handler,
        pool_upgrade_request,
        pkg_version,
        tconf):
    monkeypatch.setattr(
        NodeControlUtil, 'curr_pkg_info',
        lambda *x: (pkg_version, [APP_NAME])
    )
    monkeypatch.setattr(
        NodeControlUtil, 'get_latest_pkg_version',
        lambda *x, **y: pkg_version
    )
    pool_upgrade_request.operation[VERSION] = pkg_version.upstream.full
    pool_upgrade_request.operation[REINSTALL] = True
    pool_upgrade_handler.upgrader.get_upgrade_txn = \
        lambda predicate, reverse: \
            {TXN_PAYLOAD: {TXN_PAYLOAD_DATA: {ACTION: START}}}

    with pytest.raises(InvalidClientRequest) as e:
        pool_upgrade_handler.dynamic_validation(pool_upgrade_request, 0)
    e.match('is already scheduled')


def test_pool_upgrade_dynamic_validation_passes(
        monkeypatch,
        pool_upgrade_handler,
        pool_upgrade_request,
        pkg_version,
        tconf):
    monkeypatch.setattr(
        NodeControlUtil, 'curr_pkg_info',
        lambda *x: (pkg_version, [APP_NAME])
    )
    monkeypatch.setattr(
        NodeControlUtil, 'get_latest_pkg_version',
        lambda *x, **y: pkg_version
    )
    pool_upgrade_request.operation[VERSION] = pkg_version.upstream.full
    pool_upgrade_request.operation[REINSTALL] = True
    pool_upgrade_handler.upgrader.get_upgrade_txn = \
        lambda predicate, reverse: \
            {TXN_PAYLOAD: {TXN_PAYLOAD_DATA: {}}}
    pool_upgrade_handler.write_req_validator.validate = lambda a, b: 0
    pool_upgrade_handler.dynamic_validation(pool_upgrade_request, 0)



from indy_node.server.pool_config import PoolConfig

from indy_common.constants import CONFIG_LEDGER_ID
from indy_node.server.upgrader import Upgrader
from plenum.common.txn_util import is_forced
from plenum.server.batch_handlers.batch_request_handler import BatchRequestHandler
from plenum.server.batch_handlers.three_pc_batch import ThreePcBatch
from plenum.server.database_manager import DatabaseManager


class ConfigBatchHandler(BatchRequestHandler):
    def __init__(self, database_manager: DatabaseManager,
                 upgrader: Upgrader, pool_config: PoolConfig):
        super().__init__(database_manager, CONFIG_LEDGER_ID)
        self.upgrader = upgrader
        self.pool_config = pool_config

    def commit_batch(self, three_pc_batch: ThreePcBatch, prev_result=None):
        committed_txns = super().commit_batch(three_pc_batch, prev_result)
        for txn in committed_txns:
            # Handle POOL_UPGRADE or POOL_CONFIG transaction here
            # only in case it is not forced.
            # If it is forced then it was handled earlier
            # in applyForced method.
            if not is_forced(txn):
                self.upgrader.handleUpgradeTxn(txn)
                self.pool_config.handleConfigTxn(txn)
        return committed_txns



from indy_common.constants import NYM, NODE, ATTRIB, SCHEMA, CLAIM_DEF, DISCLO, GET_ATTR, GET_NYM, GET_TXNS, \
    GET_SCHEMA, GET_CLAIM_DEF, POOL_UPGRADE, NODE_UPGRADE, \
    POOL_CONFIG, REVOC_REG_DEF, REVOC_REG_ENTRY, \
    GET_REVOC_REG_DEF, GET_REVOC_REG, GET_REVOC_REG_DELTA, POOL_RESTART, VALIDATOR_INFO, CHANGE_KEY, AUTH_RULE, \
    GET_AUTH_RULE, AUTH_RULES, RICH_SCHEMA, RICH_SCHEMA_ENCODING, \
    RICH_SCHEMA_MAPPING, RICH_SCHEMA_CRED_DEF, JSON_LD_CONTEXT, GET_RICH_SCHEMA_OBJECT_BY_METADATA, \
    GET_RICH_SCHEMA_OBJECT_BY_ID, RICH_SCHEMA_PRES_DEF

from indy_common.transactions import IndyTransactions


def test_transactions_are_encoded():
    assert NODE == "0"
    assert NYM == "1"
    assert GET_TXNS == "3"
    assert ATTRIB == "100"
    assert SCHEMA == "101"
    assert CLAIM_DEF == "102"
    assert DISCLO == "103"
    assert GET_ATTR == "104"
    assert GET_NYM == "105"
    assert GET_SCHEMA == "107"
    assert GET_CLAIM_DEF == "108"
    assert POOL_UPGRADE == "109"
    assert NODE_UPGRADE == "110"
    assert POOL_CONFIG == "111"
    assert CHANGE_KEY == "112"

    assert REVOC_REG_DEF == "113"
    assert REVOC_REG_ENTRY == "114"
    assert GET_REVOC_REG_DEF == "115"
    assert GET_REVOC_REG == "116"
    assert GET_REVOC_REG_DELTA == "117"

    assert POOL_RESTART == "118"
    assert VALIDATOR_INFO == "119"

    assert AUTH_RULE == "120"
    assert GET_AUTH_RULE == "121"
    assert AUTH_RULES == "122"

    assert JSON_LD_CONTEXT == "200"
    assert RICH_SCHEMA == "201"
    assert RICH_SCHEMA_ENCODING == "202"
    assert RICH_SCHEMA_MAPPING == "203"
    assert RICH_SCHEMA_CRED_DEF == "204"
    assert RICH_SCHEMA_PRES_DEF == "205"
    assert GET_RICH_SCHEMA_OBJECT_BY_ID == "300"
    assert GET_RICH_SCHEMA_OBJECT_BY_METADATA == "301"


def test_transaction_enum_decoded():
    assert IndyTransactions.NODE.name == "NODE"
    assert IndyTransactions.NYM.name == "NYM"

    assert IndyTransactions.ATTRIB.name == "ATTRIB"
    assert IndyTransactions.SCHEMA.name == "SCHEMA"
    assert IndyTransactions.CLAIM_DEF.name == "CLAIM_DEF"

    assert IndyTransactions.DISCLO.name == "DISCLO"

    assert IndyTransactions.GET_ATTR.name == "GET_ATTR"
    assert IndyTransactions.GET_NYM.name == "GET_NYM"
    assert IndyTransactions.GET_TXNS.name == "GET_TXNS"
    assert IndyTransactions.GET_SCHEMA.name == "GET_SCHEMA"
    assert IndyTransactions.GET_CLAIM_DEF.name == "GET_CLAIM_DEF"

    assert IndyTransactions.POOL_UPGRADE.name == "POOL_UPGRADE"
    assert IndyTransactions.NODE_UPGRADE.name == "NODE_UPGRADE"
    assert IndyTransactions.POOL_CONFIG.name == "POOL_CONFIG"
    assert IndyTransactions.POOL_RESTART.name == "POOL_RESTART"
    assert IndyTransactions.CHANGE_KEY.name == "CHANGE_KEY"

    assert IndyTransactions.REVOC_REG_DEF.name == "REVOC_REG_DEF"
    assert IndyTransactions.REVOC_REG_ENTRY.name == "REVOC_REG_ENTRY"
    assert IndyTransactions.GET_REVOC_REG_DEF.name == "GET_REVOC_REG_DEF"
    assert IndyTransactions.GET_REVOC_REG.name == "GET_REVOC_REG"
    assert IndyTransactions.GET_REVOC_REG_DELTA.name == "GET_REVOC_REG_DELTA"

    assert IndyTransactions.VALIDATOR_INFO.name == "VALIDATOR_INFO"

    assert IndyTransactions.JSON_LD_CONTEXT.name == "JSON_LD_CONTEXT"
    assert IndyTransactions.RICH_SCHEMA.name == "RICH_SCHEMA"
    assert IndyTransactions.RICH_SCHEMA_ENCODING.name == "RICH_SCHEMA_ENCODING"
    assert IndyTransactions.RICH_SCHEMA_MAPPING.name == "RICH_SCHEMA_MAPPING"
    assert IndyTransactions.RICH_SCHEMA_CRED_DEF.name == "RICH_SCHEMA_CRED_DEF"
    assert IndyTransactions.RICH_SCHEMA_PRES_DEF.name == "RICH_SCHEMA_PRES_DEF"
    assert IndyTransactions.GET_RICH_SCHEMA_OBJECT_BY_ID.name == "GET_RICH_SCHEMA_OBJECT_BY_ID"
    assert IndyTransactions.GET_RICH_SCHEMA_OBJECT_BY_METADATA.name == "GET_RICH_SCHEMA_OBJECT_BY_METADATA"


def test_transaction_enum_encoded():
    assert IndyTransactions.NODE.value == "0"
    assert IndyTransactions.NYM.value == "1"
    assert IndyTransactions.GET_TXNS.value == "3"

    assert IndyTransactions.ATTRIB.value == "100"
    assert IndyTransactions.SCHEMA.value == "101"
    assert IndyTransactions.CLAIM_DEF.value == "102"

    assert IndyTransactions.DISCLO.value == "103"
    assert IndyTransactions.GET_ATTR.value == "104"
    assert IndyTransactions.GET_NYM.value == "105"
    assert IndyTransactions.GET_SCHEMA.value == "107"
    assert IndyTransactions.GET_CLAIM_DEF.value == "108"
    assert IndyTransactions.POOL_UPGRADE.value == "109"
    assert IndyTransactions.NODE_UPGRADE.value == "110"
    assert IndyTransactions.POOL_CONFIG.value == "111"
    assert IndyTransactions.CHANGE_KEY.value == "112"
    assert IndyTransactions.REVOC_REG_DEF.value == "113"
    assert IndyTransactions.REVOC_REG_ENTRY.value == "114"
    assert IndyTransactions.GET_REVOC_REG_DEF.value == "115"
    assert IndyTransactions.GET_REVOC_REG.value == "116"
    assert IndyTransactions.GET_REVOC_REG_DELTA.value == "117"
    assert IndyTransactions.POOL_RESTART.value == "118"
    assert IndyTransactions.VALIDATOR_INFO.value == "119"

    assert IndyTransactions.JSON_LD_CONTEXT.value == "200"
    assert IndyTransactions.RICH_SCHEMA.value == "201"
    assert IndyTransactions.RICH_SCHEMA_ENCODING.value == "202"
    assert IndyTransactions.RICH_SCHEMA_MAPPING.value == "203"
    assert IndyTransactions.RICH_SCHEMA_CRED_DEF.value == "204"
    assert IndyTransactions.RICH_SCHEMA_PRES_DEF.value == "205"
    assert IndyTransactions.GET_RICH_SCHEMA_OBJECT_BY_ID.value == "300"
    assert IndyTransactions.GET_RICH_SCHEMA_OBJECT_BY_METADATA.value == "301"


def test_get_name_from_code():
    assert IndyTransactions.get_name_from_code(IndyTransactions.NODE.value) == "NODE"
    assert IndyTransactions.get_name_from_code(IndyTransactions.NYM.value) == "NYM"

    assert IndyTransactions.get_name_from_code(IndyTransactions.ATTRIB.value) == "ATTRIB"
    assert IndyTransactions.get_name_from_code(IndyTransactions.SCHEMA.value) == "SCHEMA"
    assert IndyTransactions.get_name_from_code(IndyTransactions.CLAIM_DEF.value) == "CLAIM_DEF"

    assert IndyTransactions.get_name_from_code(IndyTransactions.DISCLO.value) == "DISCLO"
    assert IndyTransactions.get_name_from_code(IndyTransactions.GET_ATTR.value) == "GET_ATTR"
    assert IndyTransactions.get_name_from_code(IndyTransactions.GET_NYM.value) == "GET_NYM"
    assert IndyTransactions.get_name_from_code(IndyTransactions.GET_TXNS.value) == "GET_TXNS"
    assert IndyTransactions.get_name_from_code(IndyTransactions.GET_SCHEMA.value) == "GET_SCHEMA"
    assert IndyTransactions.get_name_from_code(IndyTransactions.GET_CLAIM_DEF.value) == "GET_CLAIM_DEF"
    assert IndyTransactions.get_name_from_code(IndyTransactions.POOL_UPGRADE.value) == "POOL_UPGRADE"
    assert IndyTransactions.get_name_from_code(IndyTransactions.NODE_UPGRADE.value) == "NODE_UPGRADE"
    assert IndyTransactions.get_name_from_code(IndyTransactions.POOL_CONFIG.value) == "POOL_CONFIG"
    assert IndyTransactions.get_name_from_code(IndyTransactions.POOL_RESTART.value) == "POOL_RESTART"

    assert IndyTransactions.get_name_from_code(IndyTransactions.CHANGE_KEY.value) == "CHANGE_KEY"
    assert IndyTransactions.get_name_from_code(IndyTransactions.REVOC_REG_DEF.value) == "REVOC_REG_DEF"
    assert IndyTransactions.get_name_from_code(IndyTransactions.REVOC_REG_ENTRY.value) == "REVOC_REG_ENTRY"
    assert IndyTransactions.get_name_from_code(IndyTransactions.GET_REVOC_REG_DEF.value) == "GET_REVOC_REG_DEF"
    assert IndyTransactions.get_name_from_code(IndyTransactions.GET_REVOC_REG.value) == "GET_REVOC_REG"
    assert IndyTransactions.get_name_from_code(IndyTransactions.GET_REVOC_REG_DELTA.value) == "GET_REVOC_REG_DELTA"
    assert IndyTransactions.get_name_from_code(IndyTransactions.VALIDATOR_INFO.value) == "VALIDATOR_INFO"

    assert IndyTransactions.get_name_from_code(IndyTransactions.JSON_LD_CONTEXT.value) == "JSON_LD_CONTEXT"
    assert IndyTransactions.get_name_from_code(IndyTransactions.RICH_SCHEMA.value) == "RICH_SCHEMA"
    assert IndyTransactions.get_name_from_code(
        IndyTransactions.RICH_SCHEMA_ENCODING.value) == "RICH_SCHEMA_ENCODING"
    assert IndyTransactions.get_name_from_code(
        IndyTransactions.RICH_SCHEMA_MAPPING.value) == "RICH_SCHEMA_MAPPING"
    assert IndyTransactions.get_name_from_code(
        IndyTransactions.RICH_SCHEMA_CRED_DEF.value) == "RICH_SCHEMA_CRED_DEF"
    assert IndyTransactions.get_name_from_code(
        IndyTransactions.RICH_SCHEMA_PRES_DEF.value) == "RICH_SCHEMA_PRES_DEF"
    assert IndyTransactions.get_name_from_code(
        IndyTransactions.GET_RICH_SCHEMA_OBJECT_BY_ID.value) == "GET_RICH_SCHEMA_OBJECT_BY_ID"
    assert IndyTransactions.get_name_from_code(
        IndyTransactions.GET_RICH_SCHEMA_OBJECT_BY_METADATA.value) == "GET_RICH_SCHEMA_OBJECT_BY_METADATA"

    assert IndyTransactions.get_name_from_code("some_unexpected_code") == "Unknown_transaction_type"



import json
import random

import pytest

from indy_common.authorize.auth_constraints import AuthConstraintForbidden
from indy_common.constants import RS_CONTEXT_TYPE_VALUE, JSON_LD_CONTEXT, RS_SCHEMA_TYPE_VALUE, \
    RS_MAPPING_TYPE_VALUE, RS_ENCODING_TYPE_VALUE, RS_CRED_DEF_TYPE_VALUE, RICH_SCHEMA, RICH_SCHEMA_ENCODING, \
    RICH_SCHEMA_MAPPING, RICH_SCHEMA_CRED_DEF, RICH_SCHEMA_PRES_DEF, RS_PRES_DEF_TYPE_VALUE
from indy_node.test.api.helper import sdk_write_rich_schema_object_and_check, \
    sdk_build_rich_schema_request
from indy_node.test.helper import rich_schemas_enabled_scope
from indy_node.test.rich_schema.templates import W3C_BASE_CONTEXT, RICH_SCHEMA_EX1, RICH_SCHEMA_ENCODING_EX1, \
    RICH_SCHEMA_MAPPING_EX1, RICH_SCHEMA_CRED_DEF_EX1, RICH_SCHEMA_PRES_DEF_EX1
from plenum.common.constants import TXN_PAYLOAD_METADATA_REQ_ID
from plenum.common.exceptions import RequestRejectedException, RequestNackedException
from plenum.common.types import OPERATION
from plenum.common.util import randomString
from plenum.test.helper import sdk_sign_and_submit_req, sdk_get_and_check_replies


@pytest.fixture(scope="module")
def tconf(tconf):
    with rich_schemas_enabled_scope(tconf):
        yield tconf


# The order of creation is essential as some rich schema object reference others by ID
# Encoding's id must be equal to the one used in RICH_SCHEMA_MAPPING_EX1

@pytest.mark.parametrize('txn_type, rs_type, content, rs_id',
                         [(JSON_LD_CONTEXT, RS_CONTEXT_TYPE_VALUE, W3C_BASE_CONTEXT, randomString()),
                          (RICH_SCHEMA, RS_SCHEMA_TYPE_VALUE, RICH_SCHEMA_EX1, RICH_SCHEMA_EX1['@id']),
                          (RICH_SCHEMA_ENCODING, RS_ENCODING_TYPE_VALUE, RICH_SCHEMA_ENCODING_EX1, "did:sov:1x9F8ZmxuvDqRiqqY29x6dx9oU4qwFTkPbDpWtwGbdUsrCD"),
                          (RICH_SCHEMA_MAPPING, RS_MAPPING_TYPE_VALUE, RICH_SCHEMA_MAPPING_EX1, RICH_SCHEMA_MAPPING_EX1['@id']),
                          (RICH_SCHEMA_CRED_DEF, RS_CRED_DEF_TYPE_VALUE, RICH_SCHEMA_CRED_DEF_EX1, randomString()),
                          (RICH_SCHEMA_PRES_DEF, RS_PRES_DEF_TYPE_VALUE, RICH_SCHEMA_PRES_DEF_EX1, RICH_SCHEMA_PRES_DEF_EX1['@id'])])
def test_send_rich_schema_obj(looper, sdk_pool_handle, sdk_wallet_endorser,
                              txn_type, rs_type, content, rs_id):
    # 1. check that write is successful
    request = sdk_build_rich_schema_request(looper, sdk_wallet_endorser,
                                            txn_type, rs_id=rs_id, rs_name=randomString(),
                                            rs_version='1.0', rs_type=rs_type,
                                            rs_content=json.dumps(content))
    req = sdk_sign_and_submit_req(sdk_pool_handle, sdk_wallet_endorser, request)
    rep1 = sdk_get_and_check_replies(looper, [req])

    # 2. check that sending the same request gets the same reply
    req = sdk_sign_and_submit_req(sdk_pool_handle, sdk_wallet_endorser, request)
    rep2 = sdk_get_and_check_replies(looper, [req])
    assert rep1 == rep2

    # 3. check that using a different reqId for the same request gets an error
    request2 = json.loads(request)
    request2[TXN_PAYLOAD_METADATA_REQ_ID] = random.randint(10, 1000000000)
    request2 = json.dumps(request2)

    with pytest.raises(RequestRejectedException,
                       match=str(AuthConstraintForbidden())):
        req = sdk_sign_and_submit_req(sdk_pool_handle, sdk_wallet_endorser, request2)
        sdk_get_and_check_replies(looper, [req])


@pytest.mark.parametrize('txn_type, rs_type, content, rs_id',
                         [(JSON_LD_CONTEXT, RS_CONTEXT_TYPE_VALUE, W3C_BASE_CONTEXT, randomString()),
                          (RICH_SCHEMA, RS_SCHEMA_TYPE_VALUE, RICH_SCHEMA_EX1, RICH_SCHEMA_EX1['@id']),
                          (RICH_SCHEMA_ENCODING, RS_ENCODING_TYPE_VALUE, RICH_SCHEMA_ENCODING_EX1, "did:sov:1x9F8ZmxuvDqRiqqY29x6dx9oU4qwFTkPbDpWtwGbdUsrCD"),
                          (RICH_SCHEMA_MAPPING, RS_MAPPING_TYPE_VALUE, RICH_SCHEMA_MAPPING_EX1, RICH_SCHEMA_MAPPING_EX1['@id']),
                          (RICH_SCHEMA_CRED_DEF, RS_CRED_DEF_TYPE_VALUE, RICH_SCHEMA_CRED_DEF_EX1, randomString()),
                          (RICH_SCHEMA_PRES_DEF, RS_PRES_DEF_TYPE_VALUE, RICH_SCHEMA_PRES_DEF_EX1, RICH_SCHEMA_PRES_DEF_EX1['@id'])])
@pytest.mark.parametrize('missing_field',
                         ["id", "rsName", "rsVersion", "content", "rsType"])
def test_validate_fail_missing_fields(looper, sdk_pool_handle, sdk_wallet_endorser,
                                      txn_type, rs_type, content, rs_id, missing_field):
    request = sdk_build_rich_schema_request(looper, sdk_wallet_endorser,
                                            txn_type, rs_id=rs_id, rs_name=randomString(),
                                            rs_version='1.0', rs_type=rs_type,
                                            rs_content=json.dumps(content))
    request = json.loads(request)
    request[OPERATION].pop(missing_field, None)
    request = json.dumps(request)

    with pytest.raises(RequestNackedException, match='missed fields'):
        req = sdk_sign_and_submit_req(sdk_pool_handle, sdk_wallet_endorser, request)
        sdk_get_and_check_replies(looper, [req])
